from deposit.utils.fnc_serialize import (value_to_str)

from PySide2 import (QtWidgets, QtCore, QtGui)

class DialogControl(object):
	
	changed = QtCore.Signal()
	
	def __init__(self, cmodel, user_control):
		
		self._cmodel = cmodel
		self.user_control = user_control
		self.obj_id = None
	
	def set_object(self, id):
		
		self.obj_id = id
	
	def set_value(self, value):
		
		return None
	
	def populate_lookup(self):
		
		pass
	
	def get_value(self):
		
		return None
	
class LineEdit(DialogControl, QtWidgets.QLineEdit):
	
	def __init__(self, cmodel, user_control):
		
		self.last_added = ""
		self.values = []
		self._autocompleting = True
		
		DialogControl.__init__(self, cmodel, user_control)
		QtWidgets.QLineEdit.__init__(self)
		
		self.populate_lookup()
		
		self.textChanged.connect(self.on_text_changed)
	
	def set_value(self, value):
		
		if value is None:
			value = ""
		value = str(value)
		self.blockSignals(True)
		self.setText(value)
		self.blockSignals(False)
		if value == "":
			self._autocompleting = True
		else:
			self._autocompleting = False
	
	def populate_lookup(self):
		
		self.values = [
			value_to_str(value) for value in self._cmodel.get_class_values(
				self.user_control.dclass, self.user_control.descriptor
			)
		]
	
	def get_value(self):
		
		return self.text().strip()
	
	def autocomplete(self):
		
		text = self.text()
		if not text:
			self.last_added = ""
			self._autocompleting = True
			return
		if self._autocompleting:
			for value in self.values:
				if value.lower().startswith(text.lower()):
					if value[len(text):] == self.last_added:
						self.last_added = ""
						return
					self.blockSignals(True)
					self.setText(value)
					self.setSelection(len(text), len(value))
					self.last_added = value[len(text):]
					self.blockSignals(False)
					return
		self.last_added = ""
	
	@QtCore.Slot()
	def on_text_changed(self):
		
		self.autocomplete()
		self.changed.emit()

class PlainTextEdit(DialogControl, QtWidgets.QPlainTextEdit):
	
	def __init__(self, cmodel, user_control):
		
		self.last_added = ""
		self.values = []
		self._autocompleting = True
		
		DialogControl.__init__(self, cmodel, user_control)
		QtWidgets.QPlainTextEdit.__init__(self)
		
		self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
		
		self.populate_lookup()
		
		self.textChanged.connect(self.on_text_changed)
	
	def set_value(self, value):
		
		if value is None:
			value = ""
		value = str(value)
		self.blockSignals(True)
		self.setPlainText(value)
		self.blockSignals(False)
		if value == "":
			self._autocompleting = True
		else:
			self._autocompleting = False
	
	def populate_lookup(self):
		
		self.values = [
			value_to_str(value) for value in self._cmodel.get_class_values(
				self.user_control.dclass, self.user_control.descriptor
			)
		]
	
	def get_value(self):
		
		return self.toPlainText().strip()
	
	def autocomplete(self):
		
		text = self.toPlainText()
		if not text:
			self.last_added = ""
			self._autocompleting = True
			return
		if self._autocompleting:
			for value in self.values:
				if value.lower().startswith(text.lower()):
					if value[len(text):] == self.last_added:
						self.last_added = ""
						return
					self.blockSignals(True)
					self.setPlainText(value)
					cursor = self.textCursor()
					cursor.setPosition(len(text))
					cursor.setPosition(len(value), QtGui.QTextCursor.KeepAnchor)
					self.setTextCursor(cursor)
					self.last_added = cursor.selectedText()
					self.blockSignals(False)
					return
		self.last_added = ""
	
	def set_num_rows(self, rows):
		
		doc = self.document()
		line = QtGui.QFontMetrics(doc.defaultFont()).lineSpacing()
		margins = self.contentsMargins()
		h = line * rows + (doc.documentMargin() + self.frameWidth()) * 2 + margins.top() + margins.bottom()
		if self.height() != h:
			self.setFixedHeight(h)
	
	def resize_to_fit(self):
		
		rows = max(2, self.document().size().height() + 1)
		self.set_num_rows(rows)
	
	@QtCore.Slot()
	def on_text_changed(self):
		
		self.autocomplete()
		self.resize_to_fit()
		self.changed.emit()
	
	def resizeEvent(self, event):
		
		QtWidgets.QPlainTextEdit.resizeEvent(self, event)
		self.resize_to_fit()

class ComboBox(DialogControl, QtWidgets.QComboBox):
	
	def __init__(self, cmodel, user_control):
		
		DialogControl.__init__(self, cmodel, user_control)
		QtWidgets.QComboBox.__init__(self)
		
		self.setEditable(True)
		
		self.set_value(None)
		
		self.populate_lookup()
		
		self.currentTextChanged.connect(self.on_text_changed)
	
	def set_value(self, value):
		
		self.values = [
			value_to_str(value) for value in self._cmodel.get_class_values(
				self.user_control.dclass, self.user_control.descriptor
			)
		]
		
		self.blockSignals(True)
		self.clear()
		if self.values:
			self.addItems([""] + self.values)
		if value is not None:
			value = str(value)
			if value in self.values:
				self.setCurrentIndex(self.values.index(value) + 1)
			elif self.values:
				self.setItemText(0, value)
			else:
				self.setCurrentText(value)
		self.blockSignals(False)
		
	def populate_lookup(self):
		
		self.set_value(self.currentText().strip())
		
	def get_value(self):
		
		return self.currentText().strip()
	
	@QtCore.Slot()
	def on_text_changed(self):
		
		self.changed.emit()
	
	def wheelEvent(self, event):
		
		pass

class CheckBox(DialogControl, QtWidgets.QCheckBox):
	
	def __init__(self, cmodel, user_control):
		
		DialogControl.__init__(self, cmodel, user_control)
		QtWidgets.QCheckBox.__init__(self)
		
		self.stateChanged.connect(self.on_state_changed)
	
	def set_value(self, value):
		
		try:
			value = bool(int(value))
		except:
			value = False
		self.blockSignals(True)
		self.setChecked(value)
		self.blockSignals(False)
	
	def get_value(self):
		
		return str(int(self.isChecked()))
	
	@QtCore.Slot()
	def on_state_changed(self):
		
		self.changed.emit()




from deposit.utils.fnc_serialize import (dtype_to_dict, value_to_dtype)

from PySide2 import (QtWidgets, QtCore, QtGui)

class QueryItem(object):
	
	def __init__(self, index, class_name = None, descriptor_name = None, obj_id = None, value = None, icons = None, read_only = False, obj_id_row = None):
		
		self.index = index
		self.row = None if index is None else index.row()
		self.column = None if index is None else index.column()
		self.class_name = class_name
		self.descriptor_name = descriptor_name
		self.obj_id = obj_id
		self.obj_id_row = obj_id_row
		self.value = value
		self.read_only = read_only
		self.datasource = {}
		
		self._icons = icons
		self._data = {}
		
		if (self.class_name is None) or (self.descriptor_name is None):
			self.read_only = True
		if self.value.__class__.__name__ in ["DResource", "DGeometry"]:
			self.read_only = True
		if self.obj_id != self.obj_id_row:
			self.read_only = True
	
	def to_dict(self):
		
		return dict(
			row = self.index.row(),
			column = self.index.column(),
			class_name = self.class_name,
			descriptor_name = self.descriptor_name,
			obj_id = self.obj_id,
			obj_id_row = self.obj_id_row,
			value = dtype_to_dict(self.value),
			read_only = self.read_only,
			datasource = self.datasource,
		)
	
	def from_dict(self, data):
		
		self.row = data.get("row", None)
		self.column = data.get("column", None)
		self.class_name = data.get("class_name", None)
		self.descriptor_name = data.get("descriptor_name", None)
		self.obj_id = data.get("obj_id", None)
		self.obj_id_row = data.get("obj_id_row", None)
		self.value = value_to_dtype(data.get("value", None))
		self.read_only = data.get("read_only", False)
		self.datasource = data.get("datasource", {})
		
		return self
	
	def is_object(self):
		
		return (self.obj_id is not None) and (self.descriptor_name is None)
	
	def is_datetime(self):
		
		return self.value.__class__.__name__ == "DDateTime"
	
	def is_geometry(self):
		
		return self.value.__class__.__name__ == "DGeometry"
		
	def is_resource(self):
		
		return self.value.__class__.__name__ == "DResource"
	
	def populate_data(self, role):
		
		if role in [QtCore.Qt.DisplayRole, QtCore.Qt.EditRole]:
			
			role2 = QtCore.Qt.EditRole if role == QtCore.Qt.DisplayRole else QtCore.Qt.DisplayRole
			if role2 in self._data:
				self._data[role] = self._data[role2]
			
			if self.is_object():
				self._data[role] = "ID: %d" % (self.obj_id)
				return
			
			if self.is_datetime():
				self._data[role] = self.value.isoformat
				return
			
			if self.is_geometry():
				self._data[role] = self.value.geometry_type
				return
			
			if self.is_resource():
				self._data[role] = self.value.filename
				return
			
			if self.value is None:
				self._data[role] = ""
				return
			
			self._data[role] = str(self.value)
			return
		
		if role == QtCore.Qt.DecorationRole:
			
			if self.value is None:
				self._data[role] = None
				return
			
			if self.is_object() and (self._icons is not None):
				self._data[role] = self._icons["obj"]
				return
			
			if self.is_geometry() and (self._icons is not None):
				self._data[role] = self._icons["geo"]
				return
			
			if self.is_resource():
				
				if self.value.is_image:
					
					if self.value.is_stored and (self._icons is not None):
						self._data[role] = self._icons["image"]
						return
					
					if  self._icons is not None:
						self._data[role] = self._icons["remote_image"]
						return
				
				if self.value.is_stored and (self._icons is not None):
					self._data[role] = self._icons["file"]
					return
				
				if  self._icons is not None:
					self._data[role] = self._icons["remote_file"]
					return
				
				self._data[role] = None
				return
			
			self._data[role] = None
			return
		
		if role == QtCore.Qt.UserRole:
			
			self._data[role] = self
			return
		
		if role == QtCore.Qt.BackgroundRole:
			
			if self.read_only:
				self._data[role] = QtGui.QColor(240, 240, 240, 255)
				return
			
			self._data[role] = None
			return
		
		self._data[role] = None
		return
	
	def data(self, role):
		
		if role not in self._data:
			self.populate_data(role)
		return self._data[role]
	
	def get_display_data(self):
		
		return self.data(QtCore.Qt.DisplayRole)
	
	def __repr__(self):
		
		return "QueryItem(r: %d, c: %d, Class: %s, Descriptor: %s, Object: %s, Row Object: %s, Value: %s, Read Only: %s)" % (
			self.row, self.column, self.class_name, self.descriptor_name, self.obj_id, self.obj_id_row, str(self.value), self.read_only
		)


from PySide2 import (QtCore, QtWidgets, QtGui)

class QueryList(QtWidgets.QListWidget):
	
	def __init__(self, vnavigator):
		
		QtWidgets.QListWidget.__init__(self, vnavigator)
		
		self._vnavigator = vnavigator
		
		self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
		
		self.itemActivated.connect(self.on_activated)
		self.itemSelectionChanged.connect(self.on_selected)
		
	def populate(self, queries):
		
		selected = self.get_selected()
		
		self.clear()
		
		for title in queries:
			self.addItem(QtWidgets.QListWidgetItem(title))
		
		if selected:
			for row in range(self.count()):
				item = self.item(row)
				if item.text() in selected:
					item.setSelected(True)
	
	def get_selected(self):
		
		return [item.text() for item in self.selectedItems()]
	
	@QtCore.Slot(object)
	def on_activated(self, item):
		
		self._vnavigator.on_query_activated(item.text())
	
	@QtCore.Slot()
	def on_selected(self):
		
		self._vnavigator.on_query_selected(self.get_selected())
	
	def focusInEvent(self, event):
		
		self._vnavigator.on_widget_activated()
		
		return QtWidgets.QTreeWidget.focusInEvent(self, event)
	



from deposit_gui.view.vusertools_elements.user_elements.user_labeled_element import (UserLabeledElement)
from deposit_gui.view.vusertools_elements.user_elements.user_select import (UserSelect)
from deposit_gui.view.vusertools_elements.user_elements.user_unique import (UserUnique)

class UserControl(UserLabeledElement, UserSelect):
	
	def __init__(self, stylesheet, label, dclass, descriptor):
		
		UserLabeledElement.__init__(self, stylesheet, label)
		UserSelect.__init__(self, dclass, descriptor)
	
	def to_markup(self):
		
		return "<%s [%s].[%s] style=\"%s\">%s</>" % (self.__class__.__name__, self.dclass, self.descriptor, self.stylesheet, self.label)
	
	def to_dict(self):
		
		out = dict(**UserLabeledElement.to_dict(self))
		out.update(dict(**UserSelect.to_dict(self)))
		return out

class ComboBox(UserControl):
	
	def __init__(self, stylesheet, label, dclass, descriptor):
		
		UserControl.__init__(self, stylesheet, label, dclass, descriptor)

class CheckBox(UserControl):
	
	def __init__(self, stylesheet, label, dclass, descriptor):
		
		UserControl.__init__(self, stylesheet, label, dclass, descriptor)

class LineEdit(UserControl):
	
	def __init__(self, stylesheet, label, dclass, descriptor):
		
		UserControl.__init__(self, stylesheet, label, dclass, descriptor)

class PlainTextEdit(UserControl):
	
	def __init__(self, stylesheet, label, dclass, descriptor):
		
		UserControl.__init__(self, stylesheet, label, dclass, descriptor)

class Select(UserSelect):
	
	def __init__(self, stylesheet, label, dclass, descriptor):
		
		UserSelect.__init__(self, dclass, descriptor)
	
	def to_markup(self):
		
		return "<%s [%s].[%s]/>" % (self.__class__.__name__, self.dclass, self.descriptor)
	
	def to_dict(self):
		
		return dict(
			stylesheet = "",
			label = "",
			**UserSelect.to_dict(self),
		)

class Unique(UserUnique):
	
	def __init__(self, stylesheet, label, dclass):
		
		UserUnique.__init__(self, dclass)
	
	def to_markup(self):
		
		return "<%s [%s]/>" % (self.__class__.__name__, self.dclass)
	
	def to_dict(self):
		
		return dict(
			stylesheet = "",
			label = "",
			**UserUnique.to_dict(self),
		)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from deposit_gui import __version__, __date__
import deposit_gui

from PySide2 import (QtWidgets, QtCore, QtGui)
import os

class DialogAbout(QtWidgets.QFrame):
	
	def __init__(self):
		
		QtWidgets.QFrame.__init__(self)
		
		self.setMinimumWidth(400)
		self.setLayout(QtWidgets.QVBoxLayout())
		
		content = QtWidgets.QFrame()
		content.setLayout(QtWidgets.QHBoxLayout())
		self.layout().addWidget(content)
		
		logo = QtWidgets.QLabel()
		logo.setPixmap(QtGui.QPixmap(os.path.join(os.path.dirname(deposit_gui.__file__), "res/deposit_logo.svg")))
		path = os.path.join(os.path.dirname(deposit_gui.__file__), "THIRDPARTY.TXT").replace("\\", "/")
		label = QtWidgets.QLabel('''
<h2>Deposit</h2>
<h4>Graph-based data storage and exchange</h4>
<p>Version %s (%s)</p>
<p>Copyright © <a href="mailto:peter.demjan@gmail.com">Peter Demján</a> 2013 - %s</p>
<p>&nbsp;</p>
<p>Licensed under the <a href="https://www.gnu.org/licenses/gpl-3.0.en.html">GNU General Public License</a></p>
<p><a href="https://github.com/demjanp/deposit">Home page</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href="%s">Third party libraries</a></p>
		''' % (__version__, __date__, __date__.split(".")[-1], path))
		label.setOpenExternalLinks(True)
		content.layout().addWidget(logo)
		content.layout().addWidget(label)



from PySide2 import (QtWidgets, QtCore, QtGui)

class DDialog(QtWidgets.QDialog):
	
	signal_process = QtCore.Signal(str, object)
	signal_cancel = QtCore.Signal(str, object)
	
	def __init__(self, name, view):
		
		QtWidgets.QDialog.__init__(self, view)
		
		self._name = name
		self._frame = None
		self._data = None
		self._args = []
		self._kwargs = {}
		self._button_box = None
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		
		self.accepted.connect(self.on_accepted)
		self.rejected.connect(self.on_rejected)
	
	def set_title(self, name):
		
		self.setWindowTitle(name)
	
	def set_frame(self, frame):
		
		self._frame = frame
		self.layout().addWidget(self._frame)
	
	def get_frame(self):
		
		return self._frame
	
	def set_data(self, data):
		
		self._data = data
	
	def get_data(self):
		
		return self._data
	
	def set_button_box(self, ok, cancel):
		
		if not (ok or cancel):
			return
		if ok and cancel:
			flags = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
		elif ok:
			flags = QtWidgets.QDialogButtonBox.Ok
		else:
			flags = QtWidgets.QDialogButtonBox.Cancel
		self._button_box = QtWidgets.QFrame()
		self._button_box.setLayout(QtWidgets.QVBoxLayout())
		buttons = QtWidgets.QDialogButtonBox(flags, QtCore.Qt.Horizontal)
		buttons.accepted.connect(self.accept)
		buttons.rejected.connect(self.reject)
		self._button_box.layout().addWidget(buttons)
	
	def set_enabled(self, state):
		
		if self._button_box is not None:
			self._button_box.layout().itemAt(0).widget().button(QtWidgets.QDialogButtonBox.Ok).setEnabled(state)
	
	@QtCore.Slot()
	def on_accepted(self):
		
		self.signal_process.emit(self._name, self)
	
	@QtCore.Slot()
	def on_rejected(self):
		
		self.signal_cancel.emit(self._name, self)
	



from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.view.vnavigator import VNavigator

from PySide2 import (QtCore)

class CNavigator(AbstractSubcontroller):
	
	def __init__(self, cmain) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._vnavigator = VNavigator()
		
		self._vnavigator.signal_widget_activated.connect(self.cmain.on_navigator_activated)
		self._vnavigator.signal_class_selected.connect(self.cmain.on_class_selected)
		self._vnavigator.signal_class_activated.connect(self.on_class_activated)
		self._vnavigator.signal_class_up.connect(self.on_class_up)
		self._vnavigator.signal_class_down.connect(self.on_class_down)
		self._vnavigator.signal_class_add.connect(self.on_class_add)
		self._vnavigator.signal_class_add_descriptor.connect(self.on_class_add_descriptor)
		self._vnavigator.signal_class_del_descriptor.connect(self.on_class_del_descriptor)
		self._vnavigator.signal_class_rename.connect(self.on_class_rename)
		self._vnavigator.signal_class_remove.connect(self.on_class_remove)
		
		self._vnavigator.signal_query_activated.connect(self.on_query_activated)
		self._vnavigator.signal_query_add.connect(self.on_query_add)
		self._vnavigator.signal_query_edit.connect(self.on_query_edit)
		self._vnavigator.signal_query_remove.connect(self.on_query_remove)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot(object, bool, object)
	def on_class_activated(self, cls, is_descriptor, parent_cls):
		
		if cls == "!*":
			querystr = "SELECT !*"
		elif not is_descriptor:
			querystr = "SELECT [%s].*" % (cls.name)
		else:
			return
		self.cmain.cmdiarea.add_query(querystr)
	
	@QtCore.Slot(object, object)
	def on_class_up(self, cls, cls_above):
		
		self.cmain.cmodel.switch_order(cls, cls_above)
	
	@QtCore.Slot(object, object)
	def on_class_down(self, cls, cls_below):
		
		self.cmain.cmodel.switch_order(cls, cls_below)
	
	@QtCore.Slot(object, bool, object)
	def on_class_add(self, cls, is_descriptor, parent_cls):
		
		if is_descriptor:
			cls = None
		self.cmain.cdialogs.open("AddClass", cls)
	
	@QtCore.Slot(object, bool, object)
	def on_class_add_descriptor(self, cls, is_descriptor, parent_cls):
		
		if is_descriptor:
			return
		self.cmain.cdialogs.open("AddDescriptor", cls)
	
	@QtCore.Slot(object, bool, object)
	def on_class_del_descriptor(self, cls, is_descriptor, parent_cls):
		
		if not is_descriptor:
			return
		self.cmain.cdialogs.open("DelDescriptor", cls, parent_cls)
	
	@QtCore.Slot(object, bool, object)
	def on_class_rename(self, cls, is_descriptor, parent_cls):
		
		self.cmain.cdialogs.open("RenameClass", cls, is_descriptor, parent_cls)
	
	@QtCore.Slot(object)
	def on_class_remove(self, classes):
		
		self.cmain.cdialogs.open("DelClass", classes)
	
	@QtCore.Slot(str)
	def on_query_activated(self, title):
		
		self.cmain.cmdiarea.add_query(self.cmain.cmodel.get_saved_query(title))
	
	@QtCore.Slot()
	def on_query_add(self):
		
		self.cmain.cdialogs.open("AddSavedQuery")
	
	@QtCore.Slot(str)
	def on_query_edit(self, title):
		
		self.cmain.cdialogs.open("EditSavedQuery", title)
	
	@QtCore.Slot(list)
	def on_query_remove(self, titles):
		
		self.cmain.cdialogs.open("RemoveSavedQuery", titles)
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def populate_classes(self):
		
		descriptors = self.cmain.cmodel.get_descriptor_names()
		classes = [cls for cls in self.cmain.cmodel.get_classes() if cls.name not in descriptors]
		self._vnavigator.populate_classes(classes)
	
	def populate_queries(self):
		
		self._vnavigator.populate_queries(self.cmain.cmodel.get_saved_queries())
	
	
	def deselect_all(self):
		
		self._vnavigator.deselect_all()




from PySide2 import (QtWidgets, QtCore, QtGui)

from deposit_gui import res as icon_resources
import deposit_gui

import os

class AbstractSubview(QtCore.QObject):
	
	def __init__(self, vmain = None) -> None:
		
		QtCore.QObject.__init__(self)
		
		self.vmain = vmain
		self._res_folder = None
	
	def set_res_folder(self, path):
		
		self._res_folder = path
	
	def get_icon(self, name: str) -> QtGui.QIcon:
		
		if self._res_folder is not None:
			path = os.path.join(self._res_folder, name)
			if os.path.isfile(path):
				return QtGui.QIcon(path)
		path = os.path.join(os.path.dirname(icon_resources.__file__), name)
		if os.path.isfile(path):
			return QtGui.QIcon(path)
		path = os.path.join(os.path.dirname(deposit_gui.__file__), "res", name)
		if os.path.isfile(path):
			return QtGui.QIcon(path)
		raise Exception("Could not load icon %s" % (name))



from deposit.datasource import DB
from deposit import Store

from deposit.utils.fnc_serialize import (parse_connstr)

from PySide2 import (QtWidgets, QtCore, QtGui)
from collections import defaultdict
from natsort import natsorted
import os

class DConnectTabDB(QtWidgets.QFrame):
	
	DATASOURCE = DB
	
	def __init__(self, parent):
		
		QtWidgets.QFrame.__init__(self, parent)
		
		self._can_load = False
		self._can_create = False
		
		self.datasource = self.DATASOURCE()
		
		self.parent = parent
		
		self.form = QtWidgets.QWidget()
		self.form.setLayout(QtWidgets.QFormLayout())
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(10, 10, 10, 10)
		
		self.host_combo = QtWidgets.QComboBox()
		self.host_combo.setEditable(True)
		self.host_combo.editTextChanged.connect(self.on_host_changed)
		self.dbname_combo = QtWidgets.QComboBox()
		self.dbname_combo.setEditable(True)
		self.dbname_combo.editTextChanged.connect(self.update)
		self.schema_combo = QtWidgets.QComboBox()
		self.schema_combo.setEditable(True)
		self.schema_combo.setCurrentText("public")
		self.schema_combo.editTextChanged.connect(self.update)
		self.user_edit = QtWidgets.QLineEdit("")
		self.user_edit.textChanged.connect(self.update)
		self.pass_edit = QtWidgets.QLineEdit("")
		self.pass_edit.setEchoMode(QtWidgets.QLineEdit.Password)
		self.pass_edit.textChanged.connect(self.update)
		self.identifier_combo = QtWidgets.QComboBox()
		self.identifier_combo.setEditable(True)
		self.identifier_combo.editTextChanged.connect(self.on_identifier_changed)
		self.identifier_combo.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		
		self.connect_button = QtWidgets.QPushButton(self.parent.connect_caption())
		self.connect_button.setEnabled(False)
		self.connect_button.clicked.connect(self.on_connect)
		
		self.create_button = QtWidgets.QPushButton("Create")
		self.create_button.setEnabled(False)
		self.create_button.clicked.connect(self.on_create)
		
		self.delete_button = QtWidgets.QPushButton("Delete")
		self.delete_button.setEnabled(False)
		self.delete_button.clicked.connect(self.on_delete)
		
		self.form.layout().addRow("Host[:port]:", self.host_combo)
		self.form.layout().addRow("Database:", self.dbname_combo)
		self.form.layout().addRow("Schema:", self.schema_combo)
		self.form.layout().addRow("Username:", self.user_edit)
		self.form.layout().addRow("Password:", self.pass_edit)
		self.form.layout().addRow("Identifier:", self.identifier_combo)
		
		button_container = QtWidgets.QWidget()
		button_container.setLayout(QtWidgets.QHBoxLayout())
		button_container.layout().setContentsMargins(0, 0, 0, 0)
		button_container.layout().addStretch()
		button_container.layout().addWidget(self.connect_button)
		button_container.layout().addWidget(self.create_button)
		button_container.layout().addWidget(self.delete_button)
		button_container.layout().addStretch()
		
		self.layout().addWidget(self.form)
		self.layout().addWidget(button_container)
		self.layout().addStretch()
		
		self._update_timer = QtCore.QTimer()
		self._update_timer.setSingleShot(True)
		self._update_timer.timeout.connect(self.on_update_timer)
	
	def set_recent_connections(self, data):
		# data = [[url], [identifier, connstr], ...]
		
		hosts = []
		logins = []
		dbnames = {}  # {host: [dbname, ...], ...}
		for row in data:
			if len(row) != 2:
				continue
			identifier, connstr = row
			parsed = parse_connstr(connstr)
			if not parsed:
				continue
			username = parsed["username"]
			password = parsed["password"]
			host = parsed["host"]
			dbname = parsed["dbname"]
			schema = parsed["schema"]
			if host not in dbnames:
				dbnames[host] = []
			if dbname not in dbnames[host]:
				dbnames[host].append(dbname)
			if host not in hosts:
				hosts.append(host)
				logins.append([username, password])
		self.host_combo.blockSignals(True)
		self.host_combo.addItem("")
		for i, host in enumerate(hosts):
			self.host_combo.addItem(host, [host] + logins[i] + [dbnames[host]])
		self.host_combo.blockSignals(False)
	
	def get_values(self):
		
		host = self.host_combo.currentText().strip()
		dbname = self.dbname_combo.currentText().strip()
		schema = self.schema_combo.currentText().strip()
		user = self.user_edit.text().strip()
		password = self.pass_edit.text().strip()
		identifier = self.identifier_combo.currentText().strip()
		if host and (":" not in host):
			host = "%s:5432" % (host)
		return host, dbname, schema, user, password, identifier
	
	@QtCore.Slot()
	def on_host_changed(self):
		
		data = self.host_combo.currentData()
		host = self.host_combo.currentText().strip()
		if data:
			host, username, password, dbnames = data
			_, curr_name, curr_schema, curr_user, curr_password, _ = self.get_values()
			if not (curr_user or curr_password):
				curr_user, curr_password = username, password
				self.user_edit.blockSignals(True)
				self.pass_edit.blockSignals(True)
				self.user_edit.setText(username)
				self.pass_edit.setText(password)
				self.user_edit.blockSignals(False)
				self.pass_edit.blockSignals(False)
			if dbnames:
				self.dbname_combo.blockSignals(True)
				self.dbname_combo.clear()
				self.dbname_combo.addItems(dbnames)
				if curr_name:
					self.dbname_combo.setCurrentText(curr_name)
				self.dbname_combo.blockSignals(False)
		self.update()
	
	@QtCore.Slot()
	def on_identifier_changed(self):
		
		self.update()
	
	@QtCore.Slot()
	def on_ident_load(self):
		
		self.on_update_timer()
	
	@QtCore.Slot()
	def update(self):
		
		self.connect_button.setEnabled(False)
		self.create_button.setEnabled(False)
		self.delete_button.setEnabled(False)
		
		self._update_timer.start(1000)
	
	@QtCore.Slot()
	def on_update_timer(self):
		
		host, dbname, schema, user, password, identifier = self.get_values()
		schemas = []
		identifiers = []
		if "" not in [host, dbname, schema, user, password]:
			self.datasource.set_username(user)
			self.datasource.set_password(password)
			self.datasource.set_host(host)
			self.datasource.set_database(dbname)
			self.datasource.connect()
			
			schemas = self.datasource.get_schemas()
			if schemas and not schema:
				schema = schemas[0]
			self.datasource.set_schema(schema)
			
			identifiers = natsorted(self.datasource.get_identifiers().keys())
			if identifiers and not identifier:
				identifier = identifiers[0]
			self.datasource.set_identifier(identifier)
		
		self.schema_combo.blockSignals(True)
		self.schema_combo.clear()
		self.schema_combo.addItems(schemas)
		if schema:
			self.schema_combo.setCurrentText(schema)
		self.schema_combo.blockSignals(False)
		
		self.identifier_combo.blockSignals(True)
		self.identifier_combo.clear()
		self.identifier_combo.addItems(identifiers)
		if identifier:
			self.identifier_combo.setCurrentText(identifier)
		self.identifier_combo.blockSignals(False)
		
		self._can_load = self.datasource.is_valid()
		self._can_create = (identifier != "") and self.parent.creating_enabled() and self.datasource.can_create()
		
		self.datasource.disconnect()
		
		self.connect_button.setEnabled(self._can_load)
		self.create_button.setEnabled(self._can_load or self._can_create)
		self.delete_button.setEnabled(self._can_load)
	
	@QtCore.Slot()
	def on_connect(self):
		
		if self._can_load:
			self.parent.on_connect(identifier = self.datasource.get_identifier(), connstr = self.datasource.get_connstr())
	
	@QtCore.Slot()
	def on_create(self):
		
		if self._can_load:
			reply = QtWidgets.QMessageBox.question(self, "Confirm Overwrite", "Are you sure you want to overwrite the database?<br>All data will be lost!")
			if reply != QtWidgets.QMessageBox.Yes:
				return
		
		if self._can_create:
			self.datasource.create()
			store = Store(keep_temp = True)
			folder = QtWidgets.QFileDialog.getExistingDirectory(self, caption = "Select Local Folder")
			if not folder:
				return
			store.set_local_folder(folder)
			self.datasource.save(store)
			self.parent.on_connect(identifier = self.datasource.get_identifier(), connstr = self.datasource.get_connstr(), datasource = self.datasource)
			return
	
	@QtCore.Slot()
	def on_delete(self):
		
		if self._can_load:
			reply = QtWidgets.QMessageBox.question(self, "Confirm Deletion", "Are you sure you want to delete the database?")
			if reply != QtWidgets.QMessageBox.Yes:
				return
		
		self.datasource.delete()
		self.update()




from PySide2 import (QtCore, QtWidgets, QtGui)

class ClassList(QtWidgets.QTreeWidget):
	
	def __init__(self, vnavigator):
		
		QtWidgets.QTreeWidget.__init__(self, vnavigator)
		
		self._vnavigator = vnavigator
		
		self._saved_selection = None
		
		self.setHeaderHidden(True)
		self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
		self.setExpandsOnDoubleClick(False)
		self.setIconSize(QtCore.QSize(24,24))
		self.setColumnCount(1)
		
		self.itemActivated.connect(self.on_activated)
		self.itemSelectionChanged.connect(self.on_selected)
	
	def populate(self, classes):
		
		selected = [(index.row(), index.parent().row()) for index in self.selectedIndexes()]
		
		self.clear()
		icon_obj = self._vnavigator.get_icon("object.svg")
		icon_cls = self._vnavigator.get_icon("class.svg")
		icon_descr = self._vnavigator.get_icon("descriptor.svg")
		
		items = []
		item = QtWidgets.QTreeWidgetItem()
		item.is_descriptor = False
		item.setData(0, QtCore.Qt.DisplayRole, "[classless objects]")
		item.setData(0, QtCore.Qt.UserRole, "!*")
		item.setData(0, QtCore.Qt.DecorationRole, icon_obj)
		items.append(item)
		for cls in classes:
			item = QtWidgets.QTreeWidgetItem()
			item.is_descriptor = False
			item.setData(0, QtCore.Qt.DisplayRole, cls.name)
			item.setData(0, QtCore.Qt.UserRole, cls)
			item.setData(0, QtCore.Qt.DecorationRole, icon_cls)
			subitems = []
			for descr in cls.get_descriptors(ordered = True):
				subitem = QtWidgets.QTreeWidgetItem()
				subitem.is_descriptor = True
				subitem.setData(0, QtCore.Qt.DisplayRole, descr.name)
				subitem.setData(0, QtCore.Qt.UserRole, descr)
				subitem.setData(0, QtCore.Qt.DecorationRole, icon_descr)
				subitems.append(subitem)
			for cls_sub in cls.get_subclasses(ordered = True):
				subitem = QtWidgets.QTreeWidgetItem()
				subitem.is_descriptor = False
				subitem.setData(0, QtCore.Qt.DisplayRole, cls_sub.name)
				subitem.setData(0, QtCore.Qt.UserRole, cls_sub)
				subitem.setData(0, QtCore.Qt.DecorationRole, icon_cls)
				subitems.append(subitem)
			if subitems:
				item.insertChildren(0, subitems)
			items.append(item)
		if items:
			self.insertTopLevelItems(0, items)	
		self.expandAll()
		if selected:
			if len(selected) > 1:
				self.blockSignals(True)
				for row, parent in selected[:-1]:
					parent = self.model().index(parent, 0)
					index = self.model().index(row, 0, parent)
					self.selectionModel().select(index, QtCore.QItemSelectionModel.Select)
				self.blockSignals(False)
			row, parent = selected[-1]
			parent = self.model().index(parent, 0)
			index = self.model().index(row, 0, parent)
			self.selectionModel().select(index, QtCore.QItemSelectionModel.Select)
			self.scrollTo(index)
			
	
	def get_selected(self):
		
		return [item.data(0, QtCore.Qt.UserRole) for item in self.selectedItems()]
	
	def get_selected_is_descriptor(self):
		
		selected = self.selectedItems()
		if not selected:
			return False
		return selected[0].is_descriptor
	
	def get_selected_parent(self):
		
		selected = self.selectedItems()
		if not selected:
			return None
		parent = selected[0].parent()
		if parent is None:
			return None
		return parent.data(0, QtCore.Qt.UserRole)
	
	def get_items_around_selected(self):
		
		indexes = self.selectedIndexes()
		if not indexes:
			return None, None
		
		above = self.indexAbove(indexes[0]).data(QtCore.Qt.UserRole)
		below = self.indexBelow(indexes[0]).data(QtCore.Qt.UserRole)
		
		return above, below
	
	def select_one_above(self):
		
		indexes = self.selectedIndexes()
		if not indexes:
			return
		index = self.indexAbove(indexes[0])
		self.selectionModel().select(index, QtCore.QItemSelectionModel.Select)
		self.setCurrentIndex(index)
	
	def select_one_below(self):
		
		indexes = self.selectedIndexes()
		if not indexes:
			return
		index = self.indexBelow(indexes[0])
		self.selectionModel().select(index, QtCore.QItemSelectionModel.Select)
		self.setCurrentIndex(index)
	
	def set_selected(self, names):
		
		iterator = QtWidgets.QTreeWidgetItemIterator(self)
		while iterator.value():
			item = iterator.value()
			name = item.data(0, QtCore.Qt.UserRole)
			if name in names:
				item.setSelected(True)
				names.remove(name) # select only the first occurence of name
			iterator += 1
	
	@QtCore.Slot()
	def on_selected(self):
		
		self._vnavigator.on_class_selected(self.get_selected())
	
	@QtCore.Slot(object, int)
	def on_activated(self, item, column):

		cls = item.data(0, QtCore.Qt.UserRole)
		parent = item.parent()
		if parent is not None:
			parent = parent.data(0, QtCore.Qt.UserRole)
		self._vnavigator.on_class_activated(cls, item.is_descriptor, parent)
	
	def focusInEvent(self, event):
		
		self._vnavigator.on_widget_activated()
		
		return QtWidgets.QTreeWidget.focusInEvent(self, event)
	



from deposit_gui.view.vusertools_elements.user_elements.user_element import (UserElement)

class UserSelect(UserElement):
	
	def __init__(self, dclass, descriptor):
		
		self.dclass = dclass
		self.descriptor = descriptor
		
		UserElement.__init__(self)
	
	def to_dict(self):
		
		return dict(
			dclass = self.dclass,
			descriptor = self.descriptor,
			**UserElement.to_dict(self),
		)




from PySide2 import (QtWidgets, QtCore, QtGui)

class DialogImportStore(QtWidgets.QFrame):
	
	def __init__(self, dialog, objects, classes):
		
		QtWidgets.QFrame.__init__(self)
		
		self._checkboxes = []
		
		dialog.set_title("Import Store")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		self.setMinimumWidth(256)
		
		self.setLayout(QtWidgets.QHBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		grid = QtWidgets.QFrame()
		grid.setLayout(QtWidgets.QGridLayout())
		self.layout().addWidget(grid)
		self.layout().addStretch()
		
		grid.layout().addWidget(QtWidgets.QLabel("<b>Objects:</b>"), 0, 0)
		grid.layout().addWidget(QtWidgets.QLabel("%d" % (len(objects))), 0, 1)
		grid.layout().addWidget(QtWidgets.QLabel(), 1, 0)
		grid.layout().addWidget(QtWidgets.QLabel("<b>Class</b>"), 2, 0)
		grid.layout().addWidget(QtWidgets.QLabel("<b>Unique</b>"), 2, 1)
		for idx, name in enumerate(classes):
			self._checkboxes.append(QtWidgets.QCheckBox())
			self._checkboxes[-1]._name = name
			self._checkboxes[-1].setChecked(True)
			grid.layout().addWidget(QtWidgets.QLabel(name), idx + 3, 0)
			grid.layout().addWidget(self._checkboxes[-1], idx + 3, 1)
	
	def get_unique(self):
		
		unique = set()
		for checkbox in self._checkboxes:
			if checkbox.isChecked():
				unique.add(checkbox._name)
		return unique





from deposit_gui.view.vusertools_elements.user_elements.user_element import (UserElement)

class UserLabeledElement(UserElement):
	
	def __init__(self, stylesheet, label):
		
		self.stylesheet = stylesheet
		self.label = label
		
		UserElement.__init__(self)
	
	def to_dict(self):
		
		return dict(
			stylesheet = self.stylesheet,
			label = self.label,
			**UserElement.to_dict(self),
		)




from deposit import Store
from deposit.store.dobject import DObject
from deposit.store.dclass import DClass

from deposit_gui.utils.fnc_thumbnails import (load_thumbnails, get_thumbnail)
from deposit.utils.fnc_files import (get_temp_path, sanitize_filename)

from PySide2 import (QtCore)
import shutil
import uuid
import os

class DModel(QtCore.QObject):
	
	signal_store_event = QtCore.Signal(int, list, list, object)
	# event, objects, classes, datasource
	signal_store_error = QtCore.Signal(str)
	
	EVENT_ADDED =					0x00000000
	EVENT_DELETED =					0x00000100
	EVENT_CHANGED =					0x00000200
	EVENT_SAVED =					0x00000400
	EVENT_LOADED =					0x00000800
	EVENT_LOCAL_FOLDER_CHANGED =	0x00001000
	EVENT_QUERIES_CHANGED =			0x00002000
	EVENT_USER_TOOLS_CHANGED =		0x00004000
	EVENT_SETTINGS_CHANGED =		0x00008000
	
	def __init__(self, store = None):
		
		QtCore.QObject.__init__(self)
		
		self._store = Store() if store is None else store
		self._thumbnail_cache = {}
		
		self._store.set_callback_added(self.on_added)
		self._store.set_callback_deleted(self.on_deleted)
		self._store.set_callback_changed(self.on_changed)
		self._store.set_callback_saved(self.on_saved)
		self._store.set_callback_loaded(self.on_loaded)
		self._store.set_callback_local_folder_changed(self.on_local_folder_changed)
		self._store.set_callback_queries_changed(self.on_queries_changed)
		self._store.set_callback_user_tools_changed(self.on_user_tools_changed)
		self._store.set_callback_settings_changed(self.on_settings_changed)
		self._store.set_callback_error(self.on_error)
	
	def _to_objects_and_classes(self, elements):
		
		objects = []
		classes = []
		for element in elements:
			if isinstance(element, DObject) or isinstance(element, int):
				objects.append(element)
			elif isinstance(element, DClass) or isinstance(element, str):
				classes.append(element)
		return objects, classes
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	def on_added(self, elements):
		# elements = [DObject, DClass, ...]
		
		objects, classes = self._to_objects_and_classes(elements)
		self.signal_store_event.emit(self.EVENT_ADDED, objects, classes, None)
	
	def on_deleted(self, elements):
		# elements = [obj_id, name, ...]
		
		objects, classes = self._to_objects_and_classes(elements)
		self.signal_store_event.emit(self.EVENT_DELETED, objects, classes, None)
	
	def on_changed(self, elements):
		# elements = [DObject, DClass, ...]
		
		objects, classes = self._to_objects_and_classes(elements)
		self.signal_store_event.emit(self.EVENT_CHANGED, objects, classes, None)
	
	def on_saved(self, datasource):
		
		self.signal_store_event.emit(self.EVENT_SAVED, [], [], datasource)
	
	def on_loaded(self):
		
		self.signal_store_event.emit(self.EVENT_LOADED, [], [], None)
	
	def on_local_folder_changed(self):
		
		self.signal_store_event.emit(self.EVENT_LOCAL_FOLDER_CHANGED, [], [], None)
	
	def on_queries_changed(self):
		
		self.signal_store_event.emit(self.EVENT_QUERIES_CHANGED, [], [], None)
	
	def on_user_tools_changed(self):
		
		self.signal_store_event.emit(self.EVENT_USER_TOOLS_CHANGED, [], [], None)
	
	def on_settings_changed(self):
		
		self.signal_store_event.emit(self.EVENT_SETTINGS_CHANGED, [], [], None)
	
	def on_error(self, message):
		
		self.signal_store_error.emit(message)
	
	
	# ---- General
	# ------------------------------------------------------------------------
	def clear(self) -> None:
		
		self._thumbnail_cache = {}
		self._store.clear()
	
	def get_updated_url(self, resource):
		
		return self._store.get_updated_url(resource)
	
	def set_local_folder(self, path, progress = None):
		
		self._store.set_local_folder(path, progress)
	
	def has_local_folder(self):
		
		return self._store.has_local_folder()
	
	def get_folder(self):
		
		return self._store.get_folder()
	
	def get_resource_urls(self):
		
		return self._store.get_resource_urls()
	
	def prune_resources(self):
		
		self._store.prune_resources()
	
	def open_resource(self, resource):
		# resource = DResource
		
		return self._store.open_resource(resource)
	
	def get_temp_folder(self, name, appdir = "deposit"):
		
		return get_temp_path(name, appdir = appdir)
	
	def get_temp_copy(self, resource, appdir = "deposit"):
		
		source = self.open_resource(resource)
		if source is None:
			return None
		filename = sanitize_filename(resource.filename)
		tgt_path = os.path.join(
			get_temp_path(uuid.uuid4().hex, appdir = appdir),
			filename,
		)
		target = open(tgt_path, "wb")
		shutil.copyfileobj(source, target)
		source.close()
		target.close()
		
		return tgt_path
	
	
	# ---- Create
	# ------------------------------------------------------------------------
	def add_object(self):
		
		return self._store.add_object()
	
	def add_class(self, name):
		
		return self._store.add_class(name)
	
	def add_user_tool(self, user_tool):
		
		self._store.add_user_tool(user_tool)
	
	def add_saved_query(self, title, querystr):
		
		self._store.add_saved_query(title, querystr)
	
	def add_data_row(self, 
		data: dict, 
		relations: set = set(), 
		unique: set = set(), 
		existing = {}, 
		return_added = False,
	):
		# add multiple objects with classes at once & automatically add relations 
		#	based on class relations or as specified in the relations attribute
		# data = {(Class name, Descriptor name): value, ...}
		# relations = {(Class name 1, label, Class name 2), ...}
		# unique = {Class name, ...}; always add a new object to classes 
		#	specified here, otherwise re-use objects with identical descriptors
		# existing = {Class name: Object, ...}
		#	use existing object for specified classes (i.e. just update descriptors)
		#
		# returns n_added or (n_added, added) if return_added == True
		#	added = {Class name: Object, ...}
		
		return self._store.add_data_row(data, relations, unique, existing, return_added)
	
	def import_store(self, store, unique = set(), progress = None):
		# unique = {Class name, ...}; always add a new object to classes 
		#	specified here, otherwise re-use objects with identical descriptors
		# progress = DProgress
		
		self._store.import_store(store, unique, progress)
	
	
	# ---- Read
	# ------------------------------------------------------------------------
	def get_object(self, obj_id: int):
		
		return self._store.get_object(obj_id)
	
	def get_objects(self) -> set:
		# return {DObject, ...}
		
		return self._store.get_objects()
	
	def get_object_ids(self) -> set:
		
		return self._store.get_object_ids()
	
	
	def get_class(self, name):
		
		return self._store.get_class(name)
	
	def get_classes(self, ordered = False) -> list:
		# return [DClass, ...]
		
		return self._store.get_classes(ordered)
	
	def get_class_names(self, ordered = False) -> list:
		# return [name, ...]
		
		return self._store.get_class_names(ordered)
	
	def has_class(self, name) -> bool:
		
		return self._store.has_class(name)
	
	
	def get_descriptors(self, ordered = False) -> list:
		# return [DClass, ...]
		
		return self._store.get_descriptors(ordered)
	
	def get_descriptor_names(self, ordered = False) -> list:
		# return [name, ...]
		
		return self._store.get_descriptor_names(ordered)
	
	def get_descriptor_values(self, class_name, descriptor_name, direct_only = True) -> list:
		# direct_only = if True, don't return members of subclasses
		#
		# returns [value, ...]
		
		return self._store.get_descriptor_values(class_name, descriptor_name, direct_only)
	
	
	def get_user_tools(self):
		
		return self._store.get_user_tools()
	
	
	def get_saved_query(self, title):
		
		return self._store.get_saved_query(title)
	
	def get_saved_queries(self):
		
		return self._store.get_saved_queries()
	
	
	def get_relation_labels(self) -> set:
		
		return self._store.get_relation_labels()
	
	
	def get_query(self, querystr, **kwargs):
		
		return self._store.get_query(querystr, **kwargs)
	
	
	def get_subgraph(self, objects, progress):
		
		return self._store.get_subgraph(objects, progress)
	
	
	def get_thumbnail(self, resource, size = 256):
		
		folder = self.get_folder()
		if folder is None:
			return None
		return get_thumbnail(resource, folder, self._thumbnail_cache, size)
	
	
	# ---- Update
	# ------------------------------------------------------------------------
	def switch_order(self, cls1, cls2) -> None:
		
		self._store.switch_order(cls1, cls2)
	
	def rename_class(self, cls, name):
		
		cls.rename(name)
	
	def rename_class_descriptor(self, descr, cls, name):
		
		if cls.has_descriptor(descr):
			cls.del_descriptor(descr)
			cls.set_descriptor(name)
		for obj in cls.get_members():
			obj.rename_rescriptor(descr, name)
	
	
	# ---- Delete
	# ------------------------------------------------------------------------
	def del_object(self, obj_id) -> None:
		
		self._store.del_object(obj_id)
	
	def del_class(self, name) -> None:
		
		self._store.del_class(name)
	
	def del_user_tool(self, label):
		
		self._store.del_user_tool(label)
	
	def del_class_descriptor(self, descr, cls, class_only = True):
		
		descr = self.get_class(descr)
		cls = self.get_class(cls)
		if (descr is None) or (cls is None):
			return
		cls.del_descriptor(descr)
		if class_only:
			return
		for obj in cls.get_members():
			obj.del_descriptor(descr)
	
	def del_saved_query(self, title):
		
		self._store.del_saved_query(title)
	
	
	# ---- Persistence
	# ------------------------------------------------------------------------
	def init_datasource(self, format):
		# format = Datasource or format
		
		return self._store.init_datasource(format)
	
	def has_auto_backup(self):
		
		return self._store.has_auto_backup()
	
	def set_auto_backup(self, state):
		
		self._store.set_auto_backup(state)
	
	def get_datasource(self):
		
		return self._store.get_datasource()
	
	def set_datasource(self, datasource):
		
		self._store.set_datasource(datasource)
	
	def save(self, *args, **kwargs):
		# datasource = Datasource or format
		
		return self._store.save(*args, **kwargs)
	
	def load(self, *args, **kwargs):
		# datasource = Datasource or format
		
		ret = self._store.load(*args, **kwargs)
		if not ret:
			return False
		load_thumbnails(self.get_folder(), self._thumbnail_cache)
		
		return True
	
	def is_saved(self):
		
		return self._store.is_saved()





from deposit_gui.dgui.dcactions import DCActions
from deposit_gui.view.vmdiarea_frames.external_frame import ExternalFrame
from deposit_gui.view.vmdiarea_frames.query_frame import QueryFrame

from deposit import externalsource as Externalsource
from deposit.utils.fnc_files import (as_url, sanitize_filename, get_temp_path)
from deposit.utils.fnc_serialize import (value_to_str)

from PySide2 import (QtWidgets, QtGui, QtCore)
from collections import defaultdict
from natsort import natsorted
import os

class CActions(DCActions):
	
	def __init__(self, cmain, cview) -> None:
		
		DCActions.__init__(self, cmain, cview)
		
		self._relation_label = ""
		
		self.update()
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	
	# ---- Actions
	# ------------------------------------------------------------------------
	def set_up_tool_bar(self):
		
		return {
			"Data": [
				("Connect", "Connect"),
				("Save", "Save"),
			],
			"Edit": [
				("Copy", "Copy"),
				("Paste", "Paste"),
				None,
				("AddRelation", "Add Relation"),
				("RelationName", ""),
				("RemoveRelation", "Remove Relation"),
				("ClassGraph", "Class Relations"),
				None,
				("FieldCalculator", "Field Calculator"),
				("MergeRows", "Merge Rows"),
				("MergeIdentical", "Merge Identical Rows"),
				("Duplicate", "Duplicate Row"),
			],
			"External": [
				("ImportDeposit", "Import from Deposit"),
				("ExportDeposit", "Export to Deposit"),
				None,
				("LoadExternal", "Load External Data"),
				("ImportExternal", "Import Data"),
				("ExportExternal", "Export Data"),
			]
		}
	
	def set_up_menu_bar(self):
		
		return {
			"Data": [
				("Connect", "Connect"),
				("Save", "Save"),
				("SaveAsFile", "Save As File"),
				("SaveAsPostgres", "Save As PostgreSQL"),
				None,
				("AutoBackup", "Backup database after every save"),
				None,
				("ImportDeposit", "Import from Deposit"),
				("ExportDeposit", "Export to Deposit"),
				("LoadExternal", "Load External"),
				("ExportExternal", "Export Data"),
				None,
				("SetLocalFolder", "Set Local Folder"),
				("ClearLocalFolder", "Clear Local Folder"),
				("PruneResources", "Prune Local Resources"),
			],
			"Edit": [
				("Copy", "Copy"),
				("Paste", "Paste"),
			],
			"Help": [
				("About", "About"),
				("LogFile", "Log File"),
			],
		}
	
	
	
	# implement update_[name] and on_[name] for each action
	'''
	def update_[name](self):
		
		return dict(
			caption = "Caption",
			icon = "icon.svg",
			shortcut = "Ctrl+S",
			help = "Tool tip",
			combo: list,
			checkable = True,
			checked = True,
			enabled = True,
		)
	
	def on_[name](self, state):
		
		pass
	'''
	
	def update_Connect(self):
		
		return dict(
			caption = "Connect",
			icon = "connect.svg",
			help = "Connect to Database",
			checkable = False,
			enabled = True,
		)
	
	def on_Connect(self, state):
		
		self.cmain.cdialogs.open("Connect")
	
	
	def update_Save(self):
		
		return dict(
			caption = "Save",
			icon = "save.svg",
			help = "Save",
			checkable = False,
			enabled = not self.cmain.cmodel.is_saved(),
		)
	
	def on_Save(self, state):
		
		if self.cmain.cmodel.can_save():
			self.cmain.cmodel.save()
		else:
			self.on_SaveAsFile(True)
	
	def on_SaveAsFile(self, state):
		
		path, format = self.cmain.cview.get_save_path("Save Database As", "Pickle (*.pickle);;JSON (*.json)")
		if not path:
			return
		self.cmain.cview.set_recent_dir(path)
		self.cmain.cmodel.save(path = path)
		url = as_url(path)
		self.cmain.cdialogs.open("ConfirmLoad", url)
	
	
	def update_SaveAsPostgres(self):
		
		return dict(
			help = "Save As PostgreSQL Database",
			checkable = False,
			enabled = True,
		)
	
	def on_SaveAsPostgres(self, state):
		
		self.cmain.cdialogs.open("SaveAsPostgres")
	
	
	def update_AutoBackup(self):
		
		return dict(
			help = "Backup database after every save",
			checkable = True,
			checked = self.cmain.cmodel.has_auto_backup(),
			enabled = True,
		)
	
	def on_AutoBackup(self, state):
		
		self.cmain.cmodel.set_auto_backup(state)
	
	
	def update_ImportDeposit(self):
		
		return dict(
			icon = "import.svg",
			help = "Import Deposit Database",
			checkable = False,
			enabled = True,
		)
	
	def on_ImportDeposit(self, state):
		
		self.cmain.cdialogs.open("ImportStoreSource")
	
	
	def update_ExportDeposit(self):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		return dict(
			icon = "export_deposit.svg",
			help = "Export to Deposit Database",
			checkable = False,
			enabled = isinstance(frame, QueryFrame) or (len(self.cmain.get_selected_objects()) > 0),
		)
	
	def on_ExportDeposit(self, state):
		
		objects = self.cmain.get_selected_objects()
		if not objects:
			frame = self.cmain.cmdiarea.get_current_frame()
			if isinstance(frame, QueryFrame):
				objects = set()
				for row in range(frame.get_row_count()):
					obj_id = frame.get_item(row, 0).obj_id_row
					if obj_id is not None:
						objects.add(self.cmain.cmodel.get_object(obj_id))
		if not objects:
			return
		
		store = self.cmain.cmodel.get_subgraph(objects)
		
		path, format = self.cmain.cview.get_save_path("Save Database As", "Pickle (*.pickle);;JSON (*.json)")
		if not path:
			return
		self.cmain.cview.set_recent_dir(path)
		self.cmain.cview.progress.show("Saving")
		store.save(path = path, progress = self.cmain.cview.progress)
		self.cmain.cview.progress.stop()
		url = as_url(path)
		self.cmain.cdialogs.open("ConfirmLoad", url)
	
	
	def update_SetLocalFolder(self):
		
		return dict(
			caption = "Set Local Folder",
			icon = None,
			help = "Set Local Folder",
			checkable = False,
			enabled = True,
		)
	
	def on_SetLocalFolder(self, state):
		
		folder = self.cmain.cview.get_existing_folder("Select Local Folder")
		if folder:
			self.cmain.cmodel.set_local_folder(folder)
	
	def update_ClearLocalFolder(self):
		
		return dict(
			caption = "Clear Local Folder",
			icon = "",
			help = "Clear Local Folder",
			checkable = False,
			enabled = self.cmain.cmodel.has_local_folder(),
		)
	
	def on_ClearLocalFolder(self, state):
		
		self.cmain.cmodel.set_local_folder(None)
	
	def update_PruneResources(self):
		
		return dict(
			caption = "Prune Local Resources",
			icon = "",
			help = "Prune Local Resources",
			checkable = False,
			enabled = (self.cmain.cmodel.get_folder() is not None),
		)
	
	def on_PruneResources(self, state):
		
		if self.cmain.cview.show_question(
			"Prune Local Resources",
			"Are you sure you want to delete all local resources not used in this database?"
		):
			self.cmain.cmodel.prune_resources()
	
	
	def update_AddRelation(self):
		
		return dict(
			icon = "link.svg",
			help = "Add Relation",
			checkable = False,
			enabled = len(self.cmain.get_selected_objects()) + len(self.cmain.get_selected_classes()) > 0,
		)
			
	def on_AddRelation(self, state):
		
		classes = self.cmain.get_selected_classes()
		if classes:
			self.cmain.cdialogs.open("AddRelation", list(classes), self._relation_label)
			return
		
		objects = self.cmain.get_selected_objects()
		if objects:
			self.cmain.cdialogs.open("AddRelation", list(objects), self._relation_label)
			return
	
	
	def update_RelationName(self):
		
		return dict(
			combo = natsorted(list(self.cmain.cmodel.get_relation_labels())),
			enabled = len(self.cmain.get_selected_objects()) + len(self.cmain.get_selected_classes()) > 0,
		)
	
	def on_RelationName(self, text):
		
		self._relation_label = text
	
	
	def update_RemoveRelation(self):
		
		return dict(
			icon = "unlink.svg",
			help = "Remove Relation",
			checkable = False,
			enabled = len(self.cmain.get_selected_relations()) > 0,
		)
	
	def on_RemoveRelation(self, state):
		
		relations = self.cmain.get_selected_relations()
		if relations:
			self.cmain.cdialogs.open("DelRelation", relations)
	
	
	def update_ClassGraph(self):
		
		return dict(
			icon = "classes_graph.svg",
			help = "Show Class Relations",
			checkable = False,
			enabled = True,
		)
	
	def on_ClassGraph(self, state):
		
		self.cmain.cmdiarea.add_class_graph()
	
	
	def update_FieldCalculator(self):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		return dict(
			icon = "calculator.svg",
			help = "Field Calculator",
			checkable = False,
			enabled = isinstance(frame, QueryFrame),
		)
	
	def on_FieldCalculator(self, state):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		if not isinstance(frame, QueryFrame):
			return
		rows = {}
		for row in range(frame.get_row_count()):
			obj_id = frame.get_item(row, 0).obj_id_row
			if obj_id is None:
				continue
			rows[obj_id] = {}
			for col in range(frame.get_column_count()):
				item = frame.get_item(row, col)
				rows[obj_id][(item.class_name, item.descriptor_name)] = item.value
		descriptor_name = None
		items = self.cmain.get_selected_queryitems()
		if items:
			descriptor_name = list(items)[0].descriptor_name
		if rows:
			self.cmain.cdialogs.open("FieldCalculator", rows, descriptor_name)
	
	
	def update_MergeRows(self):
		
		items = self.cmain.get_selected_queryitems()
		objects = set([item.obj_id_row for item in items])
		return dict(
			icon = "merge_objects.svg",
			help = "Merge Rows",
			checkable = False,
			enabled = len(objects) > 1,
		)
	
	def on_MergeRows(self, state):
		
		items = sorted(self.cmain.get_selected_queryitems(), key = lambda item: item.row)
		objects = []
		for item in items:
			if item.obj_id_row not in objects:
				objects.append(item.obj_id_row)
		if objects:
			self.cmain.cmodel.merge_objects(objects)
	
	
	def update_MergeIdentical(self):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		return dict(
			icon = "remove_duplicate_objects.svg",
			help = "Merge Identical Rows",
			checkable = False,
			enabled = isinstance(frame, QueryFrame),
		)
	
	def on_MergeIdentical(self, state):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		if not isinstance(frame, QueryFrame):
			return
		objects = set()
		for row in range(frame.get_row_count()):
			obj_id = frame.get_item(row, 0).obj_id_row
			if obj_id is None:
				continue
			objects.add(obj_id)
		if not objects:
			return
		if not self.cmain.cview.show_question(
			"Merge Objects", 
			"Are you sure you want to merge all Objects with identical Descriptors?"
		):
			return
		self.cmain.cmodel.merge_identical(objects)
		
	
	def update_Duplicate(self):
		
		items = self.cmain.get_selected_queryitems()
		return dict(
			icon = "split_objects.svg",
			help = "Duplicate Row",
			checkable = False,
			enabled = len(items) > 0,
		)
	
	def on_Duplicate(self, state):
		
		items = self.cmain.get_selected_queryitems()
		objects = set([item.obj_id_row for item in items])
		if objects:
			self.cmain.cmodel.duplicate(objects)
	
	
	def update_LoadExternal(self):
		
		return dict(
			icon = "load_external.svg",
			help = "Load External Data",
			checkable = False,
			enabled = True,
		)
	
	def on_LoadExternal(self, state):
		
		formats = [
			"Excel 2007+ Workbook (*.xlsx)",
			"Comma-separated Values (*.csv)",
			"ESRI Shapefile (*.shp)",
		]
		path, format = self.cmain.cview.get_load_path("Load External Data", ";;".join(formats))
		if not path:
			return
		self.cmain.cview.set_recent_dir(path)
		self.cmain.cmdiarea.add_external(as_url(path))
	
	
	def update_ImportExternal(self):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		return dict(
			icon = "import_external.svg",
			help = "Import Data",
			checkable = False,
			enabled = isinstance(frame, ExternalFrame),
		)
	
	def on_ImportExternal(self, state):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		if isinstance(frame, ExternalFrame):
			self.cmain.cdialogs.open("ImportExternal", frame)
			return
	
	
	def update_ExportExternal(self):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		return dict(
			icon = "export_external.svg",
			help = "Export Data",
			checkable = False,
			enabled = isinstance(frame, QueryFrame),
		)
	
	def on_ExportExternal(self, state):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		if not isinstance(frame, QueryFrame):
			return
		formats = [
			("Excel 2007+ Workbook (*.xlsx)", "XLSX"),
			("Comma-separated Values (*.csv)", "CSV"),
			("ESRI Shapefile (*.shp)", "SHP"),
		]
		path, format = self.cmain.cview.get_save_path(
			"Save Data As", ";;".join([format[0] for format in formats])
		)
		if not path:
			return
		if not format in dict(formats):
			return
		self.cmain.cview.set_recent_dir(path)
		source = dict(formats)[format]
		if not hasattr(Externalsource, source):
			raise Exception("Export format not supported: %s" % (source))
		externalsource = getattr(Externalsource, source)()
		ret = externalsource.save(frame.get_header, frame.get_item, 
			frame.get_row_count(), frame.get_column_count(),
			path = path
		)
		self.cmain.cview.show_information("Export Successful", "Exported to: %s" % (path))
	
	
	def update_Copy(self):
		
		items = self.cmain.get_selected_queryitems()
		return dict(
			caption = "Copy",
			icon = "copy.svg",
			shortcut = "Ctrl+C",
			help = "Copy to Clipboard",
			enabled = len(items) > 0,
		)
	
	def on_Copy(self, state):
		
		items = self.cmain.get_selected_queryitems()
		if not items:
			return
		data = defaultdict(dict)
		columns = set()
		for item in items:
			value = value_to_str(item.value)
			if value is None:
				continue
			data[item.row][item.column] = value
			columns.add(item.column)
		if (not data) or (not columns):
			return
		row_min = min(data.keys())
		row_max = max(data.keys())
		col_min = min(columns)
		col_max = max(columns)
		rows = []
		for row in range(row_min, row_max + 1):
			rows.append("\t".join([
				data.get(row, {}).get(col, "") for col in range(col_min, col_max + 1)
			]))
		rows = "\n".join(rows)
		data = QtCore.QMimeData()
		data.setData("text/plain", bytes(rows, "utf-8"))
		QtWidgets.QApplication.clipboard().setMimeData(data)
	
	def update_Paste(self):
		
		items = self.cmain.get_selected_queryitems()
		return dict(
			caption = "Paste",
			icon = "paste.svg",
			shortcut = "Ctrl+V",
			help = "Paste from Clipboard",
			enabled = len(items) == 1,
		)
	
	def on_Paste(self, state):
		
		frame = self.cmain.cmdiarea.get_current_frame()
		items = self.cmain.get_selected_queryitems()
		if len(items) != 1:
			return
		item = list(items)[0]
		if (item.obj_id is None) or (item.descriptor_name is None):
			return
		obj = self.cmain.cmodel.get_object(item.obj_id)
		
		cb = QtWidgets.QApplication.clipboard()
		mimedata = cb.mimeData()
		
		if mimedata.hasUrls():
			url = str(mimedata.urls()[0].toString())
			obj.set_resource_descriptor(item.descriptor_name, url)
			return
		
		if mimedata.hasImage():
			filename = []
			if item.class_name is not None:
				filename.append(item.class_name)
			filename.append(item.descriptor_name)
			filename.append(str(item.obj_id))
			filename = sanitize_filename("_".join(filename) + ".png")
			pixmap = QtGui.QPixmap(mimedata.imageData())
			path = os.path.join(get_temp_path(), filename)
			pixmap.save(path)
			obj.set_resource_descriptor(item.descriptor_name, as_url(path))
			return
		
		if mimedata.hasText():
			obj.set_descriptor(item.descriptor_name, mimedata.text())
			return
		
	
	def on_About(self, state):
		
		self.cmain.cdialogs.open("About")
	
	def on_LogFile(self, state):
		
		self.cmain.open_in_external(self.cmain.cview.get_logging_path())




# -*- coding: utf-8 -*-
#!/usr/bin/env python
#

from setuptools import setup, find_packages

import pathlib

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False
except ImportError:
    bdist_wheel = None

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="deposit_gui",
    version="1.4.37",
    description="GUI for Deposit: Graph database focused on scientific data collection",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/demjanp/deposit_gui",
    author="Peter Demjan",
    author_email="peter.demjan@gmail.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Database",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Microsoft :: Windows :: Windows 10",
    ],
    keywords="database, graph",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
	package_data={"":[
		"*.dll",
		"*.exe",
		"*.sh",
		"config6",
		"*.c",
		"*.i",
		"INSTALLER",
		"LICENSE",
		"METADATA",
		"*.pyd",
		"*.png",
		"*.svg",
	]},
    python_requires=">=3.10, <4",
	install_requires=[
		'deposit>=1.4.35, <1.5',
		'svgelements>=1.8.1, <2',
	],
	cmdclass={'bdist_wheel': bdist_wheel},
)



from PySide2 import (QtWidgets, QtCore, QtGui)
from collections import defaultdict

class DToolBar(QtCore.QObject):
	
	signal_triggered = QtCore.Signal(str)
	signal_combo_changed = QtCore.Signal(str, str)
	
	def __init__(self, view):
		
		QtCore.QObject.__init__(self)
		
		self.view = view
		
		self.toolbars = {} # {name: QToolBar, ...}
		self.actions = defaultdict(list)  # {name: [QAction, ...], ...}
	
	def populate(self, tools):
		# tools = {toolbar: [(name, caption), None, ...], ...}; if None: add separator
		
		for toolbar_name in tools:
			self.toolbars[toolbar_name] = self.view.addToolBar(toolbar_name)
			self.toolbars[toolbar_name].setIconSize(QtCore.QSize(36,36))
			for name in tools[toolbar_name]:
				if name is None:
					self.toolbars[toolbar_name].addSeparator()
				else:
					name, caption = name
					self.actions[name].append(QtWidgets.QAction(caption))
					self.actions[name][-1].setData(name)
					self.actions[name][-1]._toolbar_name = toolbar_name
					self.toolbars[toolbar_name].addAction(self.actions[name][-1])
			self.toolbars[toolbar_name].actionTriggered.connect(self.on_triggered)
	
	def get_names(self):
		
		return list(self.actions.keys())
	
	def update_tool(self, name, data, child_window = False, clear_combo_text = False):
		'''
		data = {
			caption: str,
			icon: str,
			shortcut: str,
			help: str,
			combo: list,
			editable: bool,
			checkable: bool,
			checked: bool,
			enabled: bool,
		}
		child_window = True if toolbar is in a child window
		'''
		
		if name not in self.actions:
			return
		if "combo" in data:
			items = data["combo"]
			if isinstance(items, list):
				collect = []
				for action in self.actions[name]:
					if not isinstance(action, ComboAction):
						toolbar_name = action._toolbar_name
						caption = action.text()
						action_ = ComboAction(name, toolbar_name, self.view)
						action_.currentTextChanged.connect(
							lambda text: self.on_combo_changed(name, text)
						)
						if caption:
							self.toolbars[toolbar_name].insertWidget(
								action, QtWidgets.QLabel(caption)
							)
						self.toolbars[toolbar_name].insertWidget(action, action_)
						self.toolbars[toolbar_name].removeAction(action)
						collect.append(action_)
						action = action_
					text = action.currentText()
					action.blockSignals(True)
					action.clear()
					action.addItems(items)
					i = action.findText(
						text, flags = QtCore.Qt.MatchExactly|QtCore.Qt.MatchCaseSensitive
					)
					if i > -1:
						action.setCurrentIndex(i)
					elif (text and not clear_combo_text):
						action.setCurrentText(text)
					if action.currentText() != text:
						self.on_combo_changed(name, action.currentText())
					if "editable" in data:
						action.setEditable(data["editable"])
					action.adjustSize()
					action.blockSignals(False)
				if collect:
					self.actions[name] = collect
		
		for action in self.actions[name]:
			if "caption" in data:
				action.setText(data["caption"])
			if ("icon" in data) and data["icon"]:
				action.setIcon(self.view.get_icon(data["icon"]))
			if "shortcut" in data:
				action.setShortcut(QtGui.QKeySequence(data["shortcut"]))
				if child_window:
					action.setShortcutContext(QtCore.Qt.WindowShortcut)
				else:
					action.setShortcutContext(QtCore.Qt.WidgetWithChildrenShortcut)
			if "help" in data:
				action.setToolTip(data["help"])
			if "checkable" in data:
				action.setCheckable(data["checkable"])
			if "checked" in data:
				action.setChecked(data["checked"])
			if "enabled" in data:
				action.setEnabled(data["enabled"])
	
	def set_toolbar_visible(self, name, state):
		
		if name not in self.toolbars:
			return
		self.toolbars[name].setVisible(state)
	
	def get_state(self, name):
		
		if name not in self.actions:
			return False
		return self.actions[name][0].isChecked()
	
	def get_combo_value(self, name):
		
		if (name in self.actions) and isinstance(self.actions[name][0], ComboAction):
			return self.actions[name][0].currentText()
	
	def set_combo_value(self, name, text):
		
		if (name in self.actions) and isinstance(self.actions[name][0], ComboAction):
			for action in self.actions[name]:
				action.blockSignals(True)
				action.setCurrentText(text)
				action.blockSignals(False)
	
	@QtCore.Slot(QtWidgets.QAction)
	def on_triggered(self, action):
		
		self.signal_triggered.emit(str(action.data()))
	
	@QtCore.Slot(str, str)
	def on_combo_changed(self, name, text):
		
		self.signal_combo_changed.emit(name, text)
	
class ComboAction(QtWidgets.QComboBox):
	
	def __init__(self, name, toolbar_name, view):
		
		QtWidgets.QComboBox.__init__(self, view)
		
		self._name = name
		self._toolbar_name = toolbar_name
		
		self.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
		self.setEditable(True)




from deposit_gui.view.vusertools_elements.dialog.dialog_form import (DialogForm)

from PySide2 import (QtCore)

class DialogSearchForm(DialogForm):
	
	signal_submit = QtCore.Signal(str)	# querystr
	
	def __init__(self, vusertools, form_tool):
		
		self.selects = []
		
		DialogForm.__init__(self, vusertools, form_tool)
		
		self.setMinimumWidth(600)
	
	def submit(self):
		
		def value_to_str(value):
			
			try:
				value = float(value)
				if value - int(value) == 0:
					value = int(value)
			except:
				value = str(value)
			if isinstance(value, str):
				return "'%s'" % (value)
			return "%s" % str(value)
		
		conditions = []
		frames, _ = self.frames()
		for frame in frames:
			value = frame.get_value()
			if value:
				conditions.append("([%s].[%s] == %s)" % (frame.dclass, frame.descriptor, value_to_str(value)))
				if [frame.dclass, frame.descriptor] not in self.selects:
					self.selects.append([frame.dclass, frame.descriptor])
		
		querystr = "SELECT %s" % (", ".join(["[%s].[%s]" % tuple(select) for select in self.selects]))
		if conditions:
			querystr += " WHERE %s" % (" and ".join(conditions))
		self.signal_submit.emit(querystr)
	
	def clear(self):
		
		frames, _ = self.frames()
		for frame in frames:
			frame.set_value("")
	
	@QtCore.Slot()
	def on_submit(self):
		
		self.submit()
		self.hide()
	
	@QtCore.Slot()
	def on_reset(self):
		
		self.clear()




from deposit_gui.view.vusertools_elements.editor.editor_frame import (EditorFrame)
from deposit_gui.view.vusertools_elements.editor.editor_group import (EditorGroup)
from deposit_gui.view.vusertools_elements.editor.editor_select import (EditorSelect)
from deposit_gui.view.vusertools_elements.editor.editor_unique import (EditorUnique)
from deposit_gui.view.vusertools_elements.editor.editor_column import (EditorColumn)
from deposit_gui.view.vusertools_elements.editor.editor_actions import (Action)
from deposit_gui.view.vusertools_elements.editor import editor_actions as EditorActions
from deposit_gui.view.vusertools_elements.user_elements.search_form import (SearchForm)
from deposit_gui.view.vusertools_elements.user_elements.entry_form import (EntryForm)
from deposit_gui.view.vusertools_elements.user_elements.user_controls import (UserControl, Select, Unique)
from deposit_gui.view.vusertools_elements.user_elements.column_break import (ColumnBreak)
from deposit_gui.view.vusertools_elements.user_elements.user_groups import (Group)

from PySide2 import (QtWidgets, QtCore)

class EditorForm(QtWidgets.QMainWindow):
	
	signal_add_tool = QtCore.Signal(object)			# form
	signal_update_tool = QtCore.Signal(str, object)	# label, form
	
	def __init__(self, vusertools, form_tool, entry):
		
		QtWidgets.QMainWindow.__init__(self, vusertools.vmain)
		
		self._vusertools = vusertools
		
		self.entry = entry
		self.form_tool = form_tool
		self.columns = []
		
		self.setWindowModality(QtCore.Qt.ApplicationModal)
		
		self.central_widget = QtWidgets.QWidget(self)
		self.central_widget.setLayout(QtWidgets.QVBoxLayout())
		self.setCentralWidget(self.central_widget)
		
		self.title_frame = QtWidgets.QFrame()
		self.title_frame.setLayout(QtWidgets.QHBoxLayout())
		self.title_frame.layout().setContentsMargins(0, 0, 0, 0)
		self.title_frame.layout().addWidget(QtWidgets.QLabel("Title:"))
		self.title_edit = QtWidgets.QLineEdit()
		self.title_frame.layout().addWidget(self.title_edit)
		self.central_widget.layout().addWidget(self.title_frame)
		
		self.controls_frame = QtWidgets.QFrame()
		self.controls_frame.setLayout(QtWidgets.QHBoxLayout())
		self.controls_frame.layout().setContentsMargins(10, 10, 10, 10)
		
		scroll_area = QtWidgets.QScrollArea()
		scroll_area.setWidgetResizable(True)
		scroll_area.setWidget(self.controls_frame)
		
		self.central_widget.layout().addWidget(scroll_area)
		
		self.selects_frame = QtWidgets.QFrame()
		self.selects_frame.setLayout(QtWidgets.QHBoxLayout())
		self.selects_frame.layout().setContentsMargins(10, 10, 10, 10)
		if not self.entry:
			self.selects_frame.layout().addWidget(QtWidgets.QLabel("SELECT"))
			self.selects_frame.layout().addStretch()
		self.central_widget.layout().addWidget(self.selects_frame)
		
		self.unique_frame = QtWidgets.QFrame()
		self.unique_frame.setLayout(QtWidgets.QHBoxLayout())
		self.unique_frame.layout().setContentsMargins(10, 10, 10, 10)
		
		button_unique = QtWidgets.QToolButton()
		button_unique.setIcon(self._vusertools.get_icon("add.svg"))
		button_unique.setIconSize(QtCore.QSize(24, 24))
		button_unique.setAutoRaise(True)
		button_unique.setToolTip("Add Class with unique objects")
		button_unique.clicked.connect(self.on_add_unique)
		
		if self.entry:
			self.unique_frame.layout().addWidget(QtWidgets.QLabel("Always add:"))
			self.unique_frame.layout().addWidget(button_unique)
			self.unique_frame.layout().addStretch()
		self.central_widget.layout().addWidget(self.unique_frame)
		
		self.setWindowTitle("Entry Form Editor")
		self.setWindowIcon(self._vusertools.get_icon("form.svg"))
		
		self.toolbar = self.addToolBar("Toolbar")
		self.toolbar.setIconSize(QtCore.QSize(36,36))
		
		for key in EditorActions.__dict__:
			action = getattr(EditorActions, key)
			if not isinstance(action, type):
				continue
			if key.startswith("_Separator"):
				self.toolbar.addSeparator()
				continue
			if (action == Action) or (not issubclass(action, Action)):
				continue
			self.toolbar.addAction(action(self))
		
		button_frame = QtWidgets.QFrame()
		button_frame.setLayout(QtWidgets.QHBoxLayout())
		button_frame.layout().setContentsMargins(5, 5, 5, 5)
		button_frame.layout().addStretch()
		button_save = QtWidgets.QPushButton("Save")
		button_save.clicked.connect(self.on_save)
		button_frame.layout().addWidget(button_save)
		button_cancel = QtWidgets.QPushButton("Cancel")
		button_cancel.clicked.connect(self.on_cancel)
		button_frame.layout().addWidget(button_cancel)
		self.central_widget.layout().addWidget(button_frame)
		
		if self.form_tool is not None:
			self.title_edit.setText(self.form_tool.label)
			for element in self.form_tool.elements:
				if issubclass(element.__class__, Select):
					self.add_select(element)
				elif isinstance(element, Unique):
					self.add_unique(element)
				elif isinstance(element, ColumnBreak):
					self.add_column()
				elif issubclass(element.__class__, UserControl):
					self.add_frame(element.__class__.__name__, element)
				elif issubclass(element.__class__, Group):
					self.add_group(element.__class__.__name__, element)
		
		if not self.columns:
			self.add_column()
		
		self.update_toolbar()
	
	def get_selected(self):
		
		for column in self.columns:
			for element in column.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
				if isinstance(element, EditorGroup) or isinstance(element, EditorFrame):
					if element.selected:
						return element
					if isinstance(element, EditorGroup):
						element = element.get_selected()
						if element is not None:
							return element
		for element in self.selects_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorSelect):
				if element.selected:
					return element
		for element in self.unique_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorUnique):
				if element.selected:
					return element
		return None
	
	def deselect_all(self):
		
		for element in self.controls_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorGroup) or isinstance(element, EditorFrame):
				element.setSelected(False)
				if isinstance(element, EditorGroup):
					element.deselect_all()
		for element in self.selects_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorSelect):
				element.setSelected(False)
		for element in self.unique_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
			if isinstance(element, EditorUnique):
				element.setSelected(False)
	
	def update_toolbar(self):
		
		selected = self.get_selected()
		for action in self.toolbar.actions():
			if action.isSeparator():
				continue
			action.update()
	
	def get_control_index(self, element):
		
		for column in self.columns:
			idx = column.layout().indexOf(element)
		return column, idx
	
	def add_column(self):
		
		self.columns.append(EditorColumn())
		self.controls_frame.layout().addWidget(self.columns[-1])
	
	def add_frame(self, element, user_control = None):
		
		selected = self.get_selected()
		if isinstance(selected, EditorGroup):
			selected.add_frame(element)
		elif isinstance(selected, EditorFrame):
			if selected.form_editor == self:
				column, idx = self.get_control_index(selected)
				column.layout().insertWidget(idx, EditorFrame(element, self, user_control))
			else:
				selected.form_editor.add_frame(element, before = selected)
		else:
			self.columns[-1].layout().addWidget(EditorFrame(element, self, user_control))
	
	def add_group(self, element, user_group = None):
		
		selected = self.get_selected()
		idx = None
		if (selected is None) or (not isinstance(selected, EditorSelect)):
			if (selected is not None) and (selected.form_editor == self):
				column, idx = self.get_control_index(selected)
		if idx is None:
			self.columns[-1].layout().addWidget(EditorGroup(element, self, user_group))
		else:
			column.layout().insertWidget(idx, EditorGroup(element, self, user_group))
	
	def add_select(self, user_select = None):
		
		selected = self.get_selected()
		if selected is not None:
			idx = self.selects_frame.layout().indexOf(selected)
		else:
			idx = self.selects_frame.layout().count() - 1
		self.selects_frame.layout().insertWidget(idx, EditorSelect(self, user_select))
	
	def add_unique(self, user_unique = None):
		
		selected = self.get_selected()
		if selected is not None:
			idx = self.unique_frame.layout().indexOf(selected)
		else:
			idx = self.unique_frame.layout().count() - 2
		self.unique_frame.layout().insertWidget(idx, EditorUnique(self, user_unique))
	
	def remove_control(self, element):
		
		column, _ = self.get_control_index(element)
		column.layout().removeWidget(element)
	
	def delete(self):
		
		selected = self.get_selected()
		if selected is None:
			return
		if isinstance(selected, EditorFrame):
			selected.form_editor.remove_control(selected)
		elif isinstance(selected, EditorGroup):
			self.remove_control(selected)
		elif isinstance(selected, EditorSelect):
			self.selects_frame.layout().removeWidget(selected)
		elif isinstance(selected, EditorUnique):
			self.unique_frame.layout().removeWidget(selected)
		selected.selected = False
		selected.setParent(None)
		self.update_toolbar()
	
	def save(self):
		
		if self.entry:
			Form = EntryForm
		else:
			Form = SearchForm
		title = self.title_edit.text()
		if title:
			form = Form(title, self._vusertools)
			for column in self.columns:
				form.elements.append(column.user_element())
				for element in column.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
					if element.parent() != column:
						continue
					if isinstance(element, EditorGroup) or isinstance(element, EditorFrame):
						form.elements.append(element.user_element())
			for element in self.selects_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
				if isinstance(element, EditorSelect):
					form.elements.append(element.user_element())
			for element in self.unique_frame.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
				if isinstance(element, EditorUnique):
					form.elements.append(element.user_element())
			if self.form_tool is None:
				self.signal_add_tool.emit(form)
			else:
				self.signal_update_tool.emit(self.form_tool.label, form)
	
	@QtCore.Slot()
	def on_selection_changed(self):
		
		self.update_toolbar()
	
	@QtCore.Slot()
	def on_add_unique(self):
		
		self.add_unique()
	
	@QtCore.Slot()
	def on_save(self):
		
		self.save()
		self.close()
	
	@QtCore.Slot()
	def on_cancel(self):
		
		self.close()




from PySide2 import (QtWidgets, QtCore, QtGui)
import webbrowser

class DClickableLogo(QtWidgets.QLabel):
	
	def __init__(self, image_path, link, *args, **kwargs):
		
		self.link = link
		
		QtWidgets.QLabel.__init__(self, *args, **kwargs)
		
		self.setPixmap(QtGui.QPixmap(image_path))
		self.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
	
	def mousePressEvent(self, event):
		
		webbrowser.open(self.link)
		QtWidgets.QLabel.mousePressEvent(self, event)




from deposit_gui.dgui.abstract_subcontroller import AbstractSubcontroller
from deposit_gui.dgui.dvdialogs import DVDialogs

from PySide2 import (QtWidgets, QtCore, QtGui)

class DCDialogs(AbstractSubcontroller):
	
	def __init__(self, cmain, cview) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._vdialogs = DVDialogs(cview._view)
		self._dialogs = {}
		
		self._vdialogs.signal_process.connect(self.on_dialog_process)
		self._vdialogs.signal_cancel.connect(self.on_dialog_cancel)
	
	def open(self, name, *args, **kwargs):
		
		if name in self._dialogs:
			self._dialogs[name].close()
			del self._dialogs[name]
		
		dialog = self._vdialogs.open(name)
		dialog._args = args
		dialog._kwargs = kwargs
		
		if hasattr(self, "set_up_%s" % (name)):
			getattr(self, "set_up_%s" % (name))(dialog, *args, **kwargs)
			if dialog._button_box is not None:
				dialog.layout().addWidget(dialog._button_box)
			dialog.adjustSize()
		
		dialog.show()
		self._dialogs[name] = dialog
	
	def close(self, name):
		
		if name in self._dialogs:
			self._dialogs[name].close()
	
	def is_open(self, name):
		
		return (name in self._dialogs) and (self._dialogs[name].isVisible())
	
	@QtCore.Slot(str, object)
	def on_dialog_process(self, name, dialog):
		
		if hasattr(self, "process_%s" % (name)):
			getattr(self, "process_%s" % (name))(dialog, *dialog._args, **dialog._kwargs)
		dialog.close()
		if name in self._dialogs:
			del self._dialogs[name]
	
	@QtCore.Slot(str, object)
	def on_dialog_cancel(self, name, dialog):
		
		if hasattr(self, "cancel_%s" % (name)):
			getattr(self, "cancel_%s" % (name))(dialog, *dialog._args, **dialog._kwargs)
		dialog.close()
		if name in self._dialogs:
			del self._dialogs[name]
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	
	# ---- Dialogs
	# ------------------------------------------------------------------------
	'''
	Implement set_up_[name], process_[name] and cancel_[name] for each dialog:
	
	def set_up_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		dialog = QtWidgets.QDialog
		
		dialog.set_title(name)
		dialog.set_frame(frame = QtWidget)
		dialog.get_frame()
		dialog.set_button_box(ok: bool, cancel: bool): set if OK and Cancel buttons are visible
		
	def process_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		process dialog after OK has been clicked
	
	def cancel_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		handle dialog after cancel has been clicked
	'''



from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_item import QueryItem

from PySide2 import (QtWidgets, QtCore)
import json

class AbstractDragModel(object):
	
	def __init__(self):
		
		self._queryframe = None
		self._proxy_model = None
		self._icons = None
	
	def supportedDragActions(self):
		
		return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction | QtCore.Qt.LinkAction | QtCore.Qt.ActionMask | QtCore.Qt.IgnoreAction
	
	def supportedDropActions(self):
		
		return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction | QtCore.Qt.LinkAction | QtCore.Qt.ActionMask | QtCore.Qt.IgnoreAction
	
	def flags(self, item):
		
		flags = QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
		
		if item is None:
			return flags
		
		if self.drag_supported(item):
			flags = flags | QtCore.Qt.ItemIsDragEnabled
		
		if self.drop_supported(item):
			flags = flags | QtCore.Qt.ItemIsDropEnabled
		
		if not item.read_only:
			flags = flags | QtCore.Qt.ItemIsEditable
		
		return flags
	
	def mimeTypes(self):
		
		return ["application/deposit", "text/uri-list", "text/plain"]
	
	def mimeData(self, indexes):
		
		paths = []
		items = []
		datasource = self._queryframe._cmodel.get_datasource().to_dict()
		datasource["_id"] = id(self._queryframe._cmodel)
		for index in indexes:
			item = index.data(QtCore.Qt.UserRole)
			if item.is_resource():
				path = self._queryframe._cmodel.get_temp_copy(item.value)
				if path is not None:
					paths.append(path)
			item.datasource = datasource.copy()
			items.append(item.to_dict())
		
		data = QtCore.QMimeData()
		if items:
			items = bytes(json.dumps(items), "utf-8")
			data.setData("application/deposit", items)
			data_cb = QtCore.QMimeData() # workaround if deposit data gets reset while dragging
			data_cb.setData("application/deposit", items)
			cb = QtWidgets.QApplication.clipboard()
			cb.setMimeData(data_cb)
		if paths:
			data.setUrls([QtCore.QUrl().fromLocalFile(path) for path in paths])
		
		return data
	
	def process_drop(self, item, data):
		
		if "application/deposit" in data.formats():
			data = json.loads(data.data("application/deposit").data().decode("utf-8"))
			items = []
			for item_data in data:
				items.append(QueryItem(None).from_dict(item_data))
			if items:
				self.on_drop_items(item, items)
				return False
		
		if data.hasUrls():
			urls = [str(url.toString()) for url in data.urls()]
			if urls:	
				self.on_drop_url(item, urls)
				return False
		
		if data.hasText():
			self.on_drop_text(item, data.text())
			return False
		
		return False
	
	def dropMimeData(self, data, action, row, column, parent):
		
		item = parent.data(QtCore.Qt.UserRole)
		if item is None:
			return False
		return self.process_drop(item, data)
	
	def drag_supported(self, item):
		# re-implement to evaluate whether item supports drag
		# item: QueryItem
		
		return True
	
	def drop_supported(self, item):
		# re-implement to evaluate whether item supports drop
		# item: QueryItem
		
		return True
	
	def on_drop_url(self, item, urls):
		# re-implement to process drop of an url
		# item: QueryItem
		
		pass
		
	def on_drop_text(self, item, text):
		# re-implement to process drop of a text
		# item: QueryItem
		
		pass
		
	def on_drop_items(self, tgt_item, items):
		# re-implement to process drop of a list of QueryItems
		# tgt_item: QueryItem
		# items: [QueryItem, ...]
		
		pass
	



from deposit_gui.view.vmdiarea_frames.abstract_mdiarea_frame import AbstractMDIAreaFrame
from deposit_gui.view.vmdiarea_frames.query_frame_elements.relation_frame import RelationFrame
from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_tab_table import QueryTabTable
from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_tab_images import (QueryTabImagesLazy, QueryTabImages)
from deposit_gui.view.vmdiarea_frames.query_frame_elements.query_tab_graph import (QueryTabGraphLazy, QueryTabGraph)

from PySide2 import (QtWidgets, QtCore, QtGui)

class QueryFrame(AbstractMDIAreaFrame, QtWidgets.QFrame):
	
	signal_query_selected = QtCore.Signal(list)		# [QueryItem, ...]
	signal_query_activated = QtCore.Signal(object)	# QueryItem
	signal_object_selected = QtCore.Signal(list)	# [DObject, ...]
	signal_relation_selected = QtCore.Signal(list)	# [(Source, Target, label), ...]
	signal_add_object = QtCore.Signal(object)	# Query
	signal_del_object = QtCore.Signal()
	signal_del_descriptor = QtCore.Signal()
	signal_edited = QtCore.Signal(object, object)	# QueryItem, value
	signal_drop_url = QtCore.Signal(object, str)	# QueryItem, url
	
	# signal_class_link = QtCore.Signal(str)				# class_name
	# signal_relation_link = QtCore.Signal(int, str, str)	# obj_id, rel_label, class_name
	# signal_relation_unlink = QtCore.Signal(int, str, str)	# obj_id, rel_label, class_name
	
	
	INITIAL_THUMBNAIL_SIZE = 128
	
	def __init__(self, query, cmodel, cview):
		
		AbstractMDIAreaFrame.__init__(self)
		QtWidgets.QFrame.__init__(self)
		
		self._query = query
		self._cmodel = cmodel
		self._cview = cview
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.layout().setSpacing(0)
		
		self.relation_frame = RelationFrame()
		self.relation_frame.signal_object_link.connect(self.on_object_link)
		self.signal_class_link = self.relation_frame.signal_class_link
		self.signal_relation_link = self.relation_frame.signal_relation_link
		self.signal_relation_unlink = self.relation_frame.signal_relation_unlink
		
		self.footer = QueryFooter(self)
		
		self.tab_table = QueryTabTable(self)
		self.tab_images = QueryTabImagesLazy(self)
		self.tab_graph = QueryTabGraphLazy(self)
		
		self.tabs = QtWidgets.QTabWidget()
		self.tabs.addTab(self.tab_table, "Table")
		self.tabs.addTab(self.tab_images, "Images")
		self.tabs.addTab(self.tab_graph, "Graph")
		
		splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
		splitter.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		self.layout().addWidget(splitter)
		
		frame_left = QtWidgets.QFrame()
		frame_left.setLayout(QtWidgets.QVBoxLayout())
		frame_left.layout().setContentsMargins(0, 0, 0, 0)
		frame_left.layout().setSpacing(0)
		frame_left.layout().addWidget(self.tabs)
		frame_left.layout().addWidget(self.footer)

		self.scroll_area = QtWidgets.QScrollArea()
		self.scroll_area.setWidgetResizable(True)
		self.scroll_area.setFrameStyle(QtWidgets.QFrame.NoFrame)
		self.scroll_area.setWidget(self.relation_frame)
		
		splitter.addWidget(frame_left)
		splitter.addWidget(self.scroll_area)
		
		self._filter_timer = QtCore.QTimer()
		self._filter_timer.setSingleShot(True)
		self._filter_timer.timeout.connect(self.on_filter_timer)
		
		self.footer.set_object_buttons_enabled(self._query.main_class != "*")
		self.footer.set_del_object_enabled(False)
		
		self.tabs.currentChanged.connect(self.on_tab_changed)
		
		self.update_count()
	
	def title(self):
		
		return self._query.querystr
	
	def icon(self):
		
		return "dep_cube.svg"
	
	def update_query(self, objects = None, classes = None):
		
		if  ((objects is None) and (classes is None)) or \
			("*" in self._query.classes) or \
			(self._query.main_class is None) or \
			set(self._query.classes).intersection([
				cls if isinstance(cls, str) else cls.name for cls in classes
			]) or self._query.objects.intersection([
				obj if isinstance(obj, int) else obj.id for obj in objects
			]):
			
			current_index = self.tabs.currentIndex()
			if current_index > 0:
				self.tabs.setCurrentIndex(0)
			
			self._query.process()
			self._cview.progress.stop()
			
			self.tab_table.update_query()
			self.relation_frame.populate()
			if isinstance(self.tab_images, QueryTabImages):
				self.tab_images.update_query(
					self.tab_table.get_images(), self.tab_table.get_item_order()
				)
			if isinstance(self.tab_graph, QueryTabGraph):
				self.tab_graph.update_query()
			
			if current_index > 0:
				self.tabs.setCurrentIndex(current_index)
	
	def select_all(self):
		
		self.get_current_tab().selectAll()
	
	def clear_selection(self):
		
		self.get_current_tab().clearSelection()
	
	def populate_tab_images(self):
		
		if isinstance(self.tab_images, QueryTabImages):
			return
		
		self.tab_images = QueryTabImages(self, self.tab_table.get_images(), self.tab_table.get_item_order(), self._cmodel)
		self.tab_images.set_thumbnail_size(self.INITIAL_THUMBNAIL_SIZE)
		self.tabs.blockSignals(True)
		self.tabs.insertTab(1, self.tab_images, "Images")
		self.tabs.removeTab(2)
		self.tabs.setCurrentIndex(1)
		self.tabs.blockSignals(False)
	
	def populate_tab_graph(self):
		
		if isinstance(self.tab_graph, QueryTabGraph):
			return
		self.tab_graph = QueryTabGraph(self, set([self._cmodel.get_object(obj_id) for obj_id in self.tab_table.get_obj_ids()]))
		self.tabs.blockSignals(True)
		self.tabs.insertTab(2, self.tab_graph, "Graph")
		self.tabs.removeTab(3)
		self.tabs.setCurrentIndex(2)
		self.tabs.blockSignals(False)
	
	def get_current_tab(self):
		
		return [self.tab_table, self.tab_images, self.tab_graph][self.tabs.currentIndex()]
	
	def update_count(self):
		
		self.footer.set_count(self.get_current_tab().get_row_count())
	
	def get_header(self, col, user_role = False):
		# pass to deposit.AbstractExternalsource to provide header data from QueryTabTable
		
		return self.tab_table._table_model.headerData(col, QtCore.Qt.Horizontal, QtCore.Qt.UserRole if user_role else QtCore.Qt.DisplayRole)
	
	def get_item(self, row, col):
		# pass to deposit.AbstractExternalsource to provide data from QueryTabTable
		
		return self.tab_table._table_model.index(row, col).data(QtCore.Qt.UserRole)
	
	def get_row_count(self):
		
		return self.tab_table.get_row_count()
	
	def get_column_count(self):
		
		return self.tab_table.get_column_count()
	
	@QtCore.Slot(int)
	def on_tab_changed(self, index):
		
		if index == 1:
			self.populate_tab_images()
		elif index == 2:
			self.populate_tab_graph()
		
		self.update_count()
		self.get_current_tab().on_selected()
	
	@QtCore.Slot(int)
	def on_zoom(self, value):
		
		self.tab_images.set_thumbnail_size(value)
	
	@QtCore.Slot()
	def on_filter(self):
		
		self._filter_timer.start(1000)
	
	@QtCore.Slot()
	def on_filter_timer(self):
		
		self.tab_table.apply_filter(self.footer.get_filter_text())
		if isinstance(self.tab_images, QueryTabImages):
			self.tab_images.apply_filter(self.tab_table.get_item_order())
		if isinstance(self.tab_graph, QueryTabGraph):
			self.tab_graph.apply_filter(set([self._cmodel.get_object(obj_id) for obj_id in self.tab_table.get_obj_ids()]))
		self.update_count()
	
	@QtCore.Slot()
	def on_sorted(self):
		
		self.tab_images.sort(self.tab_table.get_item_order())
	
	@QtCore.Slot()
	def on_add_object(self):
		
		self.signal_add_object.emit(self._query)
	
	@QtCore.Slot()
	def on_del_object(self):
		
		self.signal_del_object.emit()
	
	@QtCore.Slot()
	def on_del_descriptor(self):
		
		self.signal_del_descriptor.emit()
	
	@QtCore.Slot(int)
	def on_to_object(self, obj_id):
		
		if obj_id is None:
			return
		self.tabs.setCurrentIndex(0)
		self.tab_table.select_object(obj_id)
	
	@QtCore.Slot(int)
	def on_object_link(self, obj_id):
		
		self.on_to_object(obj_id)
	
	def on_query_activated(self, item):
		
		self.signal_query_activated.emit(item)
	
	def on_query_selected(self, items):
		
		self.signal_query_selected.emit(items)
		
		has_descriptor = False
		for item in items:
			if (item.obj_id is not None) and (item.value is not None):
				has_descriptor = True
				break
		found = False
		for item in items:
			if item.obj_id is not None:
				self.relation_frame.populate(self._cmodel.get_object(item.obj_id))
				found = True
				break
		if not found:
			self.relation_frame.populate()
		
		self.footer.set_del_descriptor_enabled(has_descriptor)
	
	def on_object_activated(self, obj_id):
		
		self.on_to_object(obj_id)
	
	def on_object_selected(self, object_ids):
		
		objects = [self._cmodel.get_object(obj_id) for obj_id in object_ids]
		obj = None
		if len(objects) == 1:
			obj = objects[0]
		self.relation_frame.populate()
		self.signal_object_selected.emit(objects)
	
	def on_relation_selected(self, relations):
		
		self.signal_relation_selected.emit([(self._cmodel.get_object(source), self._cmodel.get_object(target), label) for source, target, label in relations])
	
	def on_selected_rows(self, row_items):
		
		self.footer.set_del_object_enabled(len(row_items) > 0)
	
	def on_edited(self, item, value):
		
		self.signal_edited.emit(item, value)
	
	def on_drop_url(self, item, url):
		
		self.signal_drop_url.emit(item, url)
	
	def on_deactivate(self):
		
		self.tab_table.clearSelection()
		self.tab_images.clearSelection()
		self.tab_graph.deselect_all()
	
	def on_close(self):
		
		self._filter_timer.stop()
		self.tab_table.on_close()
		self.tab_images.on_close()
		self.tab_graph.on_close()

class QueryFooter(QtWidgets.QFrame):

	def __init__(self, queryframe):

		QtWidgets.QFrame.__init__(self)
		
		self._queryframe = queryframe
		self._count_text = None
		
		self.setFrameShape(QtWidgets.QFrame.StyledPanel)
		self.setFrameShadow(QtWidgets.QFrame.Raised)

		self.setLayout(QtWidgets.QGridLayout())
		self.layout().setContentsMargins(5, 0, 0, 0)
		self.layout().setSpacing(0)
		

		self.add_object_button = QtWidgets.QToolButton()
		self.add_object_button.setIcon(self._queryframe.get_icon("add_object.svg"))
		self.add_object_button.setIconSize(QtCore.QSize(24, 24))
		self.add_object_button.setAutoRaise(True)
		self.add_object_button.setToolTip("Add Object")
		self.add_object_button.setContentsMargins(0, 0, 0, 0)
		self.add_object_button.clicked.connect(self._queryframe.on_add_object)
		self.layout().addWidget(self.add_object_button, 0, 0)
		
		self.del_object_button = QtWidgets.QToolButton()
		self.del_object_button.setIcon(self._queryframe.get_icon("remove_object.svg"))
		self.del_object_button.setIconSize(QtCore.QSize(24, 24))
		self.del_object_button.setAutoRaise(True)
		self.del_object_button.setToolTip("Remove Object")
		self.del_object_button.setContentsMargins(0, 0, 0, 0)
		self.del_object_button.clicked.connect(self._queryframe.on_del_object)
		self.layout().addWidget(self.del_object_button, 0, 1)
		
		self.del_descriptor_button = QtWidgets.QToolButton()
		self.del_descriptor_button.setIcon(self._queryframe.get_icon("remove_descriptor.svg"))
		self.del_descriptor_button.setIconSize(QtCore.QSize(24, 24))
		self.del_descriptor_button.setAutoRaise(True)
		self.del_descriptor_button.setToolTip("Remove Descriptor")
		self.del_descriptor_button.setContentsMargins(0, 0, 5, 0)
		self.del_descriptor_button.clicked.connect(self._queryframe.on_del_descriptor)
		self.layout().addWidget(self.del_descriptor_button, 0, 2)
		
		filter_label = QtWidgets.QLabel("Filter:")
		filter_label.setContentsMargins(5, 0, 0, 0)
		self.layout().addWidget(filter_label, 0, 3)

		self._filter_edit = QtWidgets.QLineEdit()
		self._filter_edit.setContentsMargins(5, 0, 5, 0)
		self._filter_edit.textEdited.connect(self._queryframe.on_filter)
		self.layout().addWidget(self._filter_edit, 0, 4)

		self._count_label = QtWidgets.QLabel("Found: %s")
		self._count_label.setContentsMargins(0, 0, 5, 0)
		self.layout().addWidget(self._count_label, 0, 5)
		
		self._count_text = self._count_label.text()
	
	def get_filter_text(self):
		
		return self._filter_edit.text()
	
	def set_count(self, count):

		self._count_label.setText(self._count_text % (count))
	
	def set_object_buttons_enabled(self, state):

		self.add_object_button.setVisible(state)
		self.del_object_button.setVisible(state)
	
	def set_del_object_enabled(self, state):
		
		self.del_object_button.setEnabled(state)
	
	def set_del_descriptor_enabled(self, state):
		
		self.del_descriptor_button.setEnabled(state)




from deposit_gui.view.vusertools_elements.dialog.dialog_frame import (DialogFrame)

from PySide2 import (QtWidgets, QtCore)
from natsort import (natsorted)

class DialogGroup(QtWidgets.QGroupBox):
	
	entry_removed = QtCore.Signal(list) # [obj_id, ...]
	
	def __init__(self, cmodel, user_group):
		# user_group = Group
		
		self._cmodel = cmodel
		self.user_group = user_group
		self.controls_frame = None
		
		QtWidgets.QGroupBox.__init__(self, self.user_group.label)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(0, 0, 0, 0)
		self.controls_frame = QtWidgets.QFrame()
		self.controls_frame.setLayout(QtWidgets.QVBoxLayout())
		self.controls_frame.layout().setContentsMargins(10, 10, 10, 10)
		self.layout().addWidget(self.controls_frame)
		
		self.lookup_combo = LookupCombo()
		self.lookup_combo_search = {} # {index: {column: value, ...}}
		lookup_button = QtWidgets.QPushButton("Fill")
		lookup_button.clicked.connect(self.on_lookup_combo)
		
		lookup_frame = QtWidgets.QFrame()
		lookup_frame.setLayout(QtWidgets.QHBoxLayout())
		lookup_frame.layout().setContentsMargins(0, 0, 0, 0)
		lookup_frame.layout().addStretch()
		lookup_frame.layout().addWidget(self.lookup_combo)
		lookup_frame.layout().addWidget(lookup_button)
		self.controls_frame.layout().addWidget(lookup_frame)
		
		for member in self.user_group.members:
			self.add_frame(member)
		
		button_frame = QtWidgets.QWidget()
		button_frame.setLayout(QtWidgets.QHBoxLayout())
		button_frame.layout().setContentsMargins(0, 0, 0, 0)
		button_remove = QtWidgets.QPushButton("Remove")
		button_remove.clicked.connect(self.on_remove)
		button_frame.layout().addStretch()
		button_frame.layout().addWidget(button_remove)
		self.controls_frame.layout().addWidget(button_frame)
		
		self.populate_lookup()
		
		if self.user_group.stylesheet:
			self.setStyleSheet(self.user_group.stylesheet)
	
	def add_frame(self, user_control):
		
		frame = DialogFrame(self._cmodel, user_control)
		frame.ctrl.changed.connect(self.on_ctrl_changed)
		self.controls_frame.layout().addWidget(frame)
	
	def populate_lookup(self):
		
		def str_or_empty(value):
			
			if value is None:
				return ""
			return str(value)
		
		self.lookup_combo.clear()
		self.lookup_combo_search = {} # {index: {column: value, ...}}
		columns = [(member.dclass, member.descriptor) for member in self.user_group.members]
		query = self._cmodel.get_query("SELECT %s" % (", ".join(["[%s].[%s]" % (column) for column in columns])), silent = True)
		rows = []
		for row in range(len(query)):
			data = dict([("[%s].[%s]" % (cls, descr), str_or_empty(query[row, cls, descr][1])) for cls, descr in columns])
			label = ", ".join(list(data.values()))
			row = [label, data]
			if row not in rows:
				rows.append(row)
		
		rows = natsorted(rows, key = lambda row: row[0])
		index = 0
		for label, data in rows:
			self.lookup_combo_search[index] = data
			index += 1
			self.lookup_combo.addItem(label, data)
	
	def remove(self):
		
		obj_ids = set([])
		for frame in self.frames():
			obj_id = frame.ctrl.obj_id
			if obj_id is not None:
				obj_ids.add(obj_id)
			frame.set_value("", None)
		return list(obj_ids)
	
	def set_data(self, data):
		# data = {Class.Descriptor: value, ...}
		
		for frame in self.frames():
			column = "[%s].[%s]" % (frame.user_control.dclass, frame.user_control.descriptor)
			if column in data:
				frame.set_value(data[column])
	
	def update_lookup_combo(self):
		
		def _find_index(data):
			
			for index in self.lookup_combo_search:
				found = True
				for column in data:
					if column not in self.lookup_combo_search[index]:
						found = False
						break
					if self.lookup_combo_search[index][column] != data[column]:
						found = False
						break
				if found:
					return index
			return None
		
		data = {}
		for frame in self.frames():
			value = frame.get_value()
			if value != "":
				column = "[%s].[%s]" % (frame.user_control.dclass, frame.user_control.descriptor)
				data[column] = value
		if not data:
			return
		index = _find_index(data)
		if index is None:
			return
		self.lookup_combo.setCurrentIndex(index)
	
	def frames(self):
		
		return list(self.controls_frame.findChildren(DialogFrame, options = QtCore.Qt.FindDirectChildrenOnly))
	
	@QtCore.Slot()
	def on_ctrl_changed(self):
		
		self.update_lookup_combo()
	
	@QtCore.Slot()
	def on_lookup_combo(self):
		
		data = self.lookup_combo.currentData()
		if data is None:
			return
		self.set_data(data)
	
	@QtCore.Slot()
	def on_remove(self):
		
		obj_ids = self.remove()
		self.entry_removed.emit(obj_ids)

class LookupCombo(QtWidgets.QComboBox):
	
	def wheelEvent(self, event):
		
		pass




from deposit_gui.view.vusertools_elements.user_elements.query import (Query)

from PySide2 import (QtWidgets, QtCore)

class EditorQuery(QtWidgets.QDialog):
	
	signal_add_tool = QtCore.Signal(object)			# query
	signal_update_tool = QtCore.Signal(str, object)	# label, query
	
	def __init__(self, vusertools, query_tool):
		
		QtWidgets.QDialog.__init__(self, vusertools.vmain)
		
		self._vusertools = vusertools
		self.query_tool = query_tool
		
		self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		self.setWindowTitle("Edit Query Tool")
		
		self.setMinimumWidth(600)
		self.setModal(True)
		self.setLayout(QtWidgets.QVBoxLayout())
		
		self.title_edit = QtWidgets.QLineEdit()
		self.query_edit = QtWidgets.QPlainTextEdit()
		form = QtWidgets.QFrame()
		form.setLayout(QtWidgets.QFormLayout())
		form.layout().setContentsMargins(0, 0, 0, 0)
		form.layout().addRow("Title:", self.title_edit)
		form.layout().addRow("Query:", self.query_edit)
		self.layout().addWidget(form)
		
		self.buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, QtCore.Qt.Horizontal)
		self.buttonBox.accepted.connect(self.accept)
		self.buttonBox.rejected.connect(self.reject)
		self.layout().addWidget(self.buttonBox)
		
		self.finished.connect(self.on_finished)
		
		if self.query_tool is not None:
			self.title_edit.setText(self.query_tool.label)
			self.query_edit.setPlainText(self.query_tool.value)
	
	def save(self):
		
		title = self.title_edit.text()
		querystr = self.query_edit.toPlainText()
		if title and querystr:
			query = Query(title, querystr, self._vusertools)
			if self.query_tool is None:
				self.signal_add_tool.emit(query)
			else:
				self.signal_update_tool.emit(self.query_tool.label, query)
	
	@QtCore.Slot(object)
	def on_finished(self, code):
		
		if code == QtWidgets.QDialog.Accepted:
			self.save()


from deposit_gui.controller.controller import Controller
from deposit.utils.fnc_files import (clear_temp_dir)

from PySide2 import (QtWidgets)
import sys

class DGUIMain(object):
	
	def __init__(self, store: object = None) -> None:
		
		app = QtWidgets.QApplication(sys.argv)
#		app.setStyle("Fusion")
		
		self.controller = Controller(store)
		self.controller.cview.show()
		self.controller.cdialogs.open("Connect")
		
		app.exec_()
	
	def __del__(self):
		
		clear_temp_dir()




from PySide2 import (QtWidgets, QtCore, QtGui)

class Action(QtWidgets.QAction):
	
	def __init__(self, form_editor):
		
		self.form_editor = form_editor
		
		QtWidgets.QAction.__init__(self, self.form_editor._vusertools.vmain)
		
		self.setText(self.name())
		icon = self.icon()
		if icon:
			icon = self.form_editor._vusertools.get_icon(icon)
			if icon:
				self.setIcon(icon)
		self.setToolTip(self.help())
		self.setCheckable(self.checkable())
		self.triggered.connect(self.on_triggered)
	
	def name(self):
		
		return self.__class__.__name__
	
	def icon(self):
		
		return ".svg"
	
	def help(self):
		
		return self.name()
	
	def enabled(self):
		
		return True
	
	def checkable(self):
		
		return False
	
	def checked(self):
		
		return False
	
	def visible(self):
		
		return True
	
	def update(self):
		
		self.setEnabled(self.enabled())
		self.setVisible(self.visible())
		self.setChecked(self.checked())
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		pass

class LineEdit(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "line_edit.svg"
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_frame(self.__class__.__name__)

class PlainTextEdit(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "plain_text_edit.svg"
	
	def enabled(self):
		
		return self.form_editor.entry
	
	def visible(self):
		
		return self.form_editor.entry
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_frame(self.__class__.__name__)

class ComboBox(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "combo_box.svg"
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_frame(self.__class__.__name__)

class CheckBox(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "check_box.svg"
	
	def enabled(self):
		
		return self.form_editor.entry
	
	def visible(self):
		
		return self.form_editor.entry
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_frame(self.__class__.__name__)

class Group(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "group_box.svg"
	
	def enabled(self):
		
		return self.form_editor.entry
		
	def visible(self):
		
		return self.form_editor.entry
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_group(self.__class__.__name__)

class MultiGroup(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "multi_group_box.svg"
	
	def enabled(self):
		
		return self.form_editor.entry
		
	def visible(self):
		
		return self.form_editor.entry
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_group(self.__class__.__name__)

class ColumnBreak(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "column_break.svg"
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_column()

class Select(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "select.svg"
	
	def enabled(self):
		
		return not self.form_editor.entry
		
	def visible(self):
		
		return not self.form_editor.entry
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.add_select()

class _Separator1(): pass

class Bold(Action):
	
	def icon(self):
		
		return "bold.svg"
	
	def enabled(self):
		
		return self.form_editor.get_selected().__class__.__name__ in ["EditorFrame", "EditorGroup"]
	
	def checkable(self):
		
		return True
	
	def checked(self):
		
		if self.form_editor.get_selected().__class__.__name__ not in ["EditorFrame", "EditorGroup"]:
			return False
		return self.form_editor.get_selected().bold
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.get_selected().setBold(state)

class OrderUp(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "up_small_black.svg"
	
	def enabled(self):
		
		return self.form_editor.get_selected() is not None
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		pass

class OrderDown(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "down_small_black.svg"
	
	def enabled(self):
		
		return self.form_editor.get_selected() is not None
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		pass

class Delete(Action):
	
	def __init__(self, form_editor):
		
		Action.__init__(self, form_editor)
	
	def icon(self):
		
		return "delete.svg"
	
	def enabled(self):
		
		return self.form_editor.get_selected() is not None
	
	@QtCore.Slot(bool)
	def on_triggered(self, state):
		
		self.form_editor.delete()




from deposit_gui.dgui.dconnect_tabs.dconnect_tab_db import DConnectTabDB

from deposit.datasource import DBRel

from PySide2 import (QtWidgets, QtCore, QtGui)

class DConnectTabDBRel(DConnectTabDB):
	
	DATASOURCE = DBRel



from PySide2 import (QtWidgets, QtCore)
import os
import logging
from pathlib import Path
from logging.handlers import RotatingFileHandler

class DLogging(QtCore.QObject):
	
	logged = QtCore.Signal(str)
	
	def __init__(self, app_name, version):
		
		QtCore.QObject.__init__(self)
		
		self.log_path = ""
		self.version = version
		
		self._append_data = []
		
		path = os.path.join(str(Path.home()), app_name)
		if not os.path.isdir(path):
			os.mkdir(path)
		self.log_path = os.path.join(path, "%s.log" % (app_name))
		log_handler = RotatingFileHandler(self.log_path, mode = "a", maxBytes = 100*1024, backupCount = 2, encoding = None, delay = 0)
		log_handler.setFormatter(logging.Formatter(fmt = "%(asctime)s v%(version)s %(message)s", datefmt = "%Y-%m-%d %H:%M:%S"))
		log_handler.setLevel(logging.INFO)
		self.logger = logging.getLogger("root")
		self.logger.setLevel(logging.INFO)
		self.logger.addHandler(log_handler)
	
	def get_log_path(self):
		
		return self.log_path
	
	def append(self, text):
		
		text = str(text).strip().strip("\n").strip("\r")
		if not text:
			return
		self.logger.info(text, extra = {"version": self.version})
		self.logged.emit(text)



from deposit_gui.view.vusertools_elements.dialog.dialog_column import (DialogColumn)
from deposit_gui.view.vusertools_elements.dialog.dialog_frame import (DialogFrame)
from deposit_gui.view.vusertools_elements.dialog.dialog_group import (DialogGroup)
from deposit_gui.view.vusertools_elements.dialog.dialog_multi_group import (DialogMultiGroup)
from deposit_gui.view.vusertools_elements.user_elements.user_groups import (Group, MultiGroup)
from deposit_gui.view.vusertools_elements.user_elements.user_controls import (UserControl, Select, Unique)
from deposit_gui.view.vusertools_elements.user_elements.column_break import (ColumnBreak)

from PySide2 import (QtWidgets, QtCore)

class DialogForm(QtWidgets.QDialog):

	def __init__(self, vusertools, form_tool):
		
		QtWidgets.QDialog.__init__(self, vusertools.vmain)
		
		self._vusertools = vusertools
		
		self.form_tool = form_tool
		self.buttonBox = None
		self.columns = []  # [DialogColumn(), ...]
		self.selects = []  # [[class, descriptor], ...]
		self.unique = set()  # {class, ...}
		
		self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		
		self.setWindowTitle(self.form_tool.label)
		self.setModal(False)
		
		self.setLayout(QtWidgets.QVBoxLayout())
		
		self.layout().setContentsMargins(0, 0, 0, 0)
		
		self.controls_frame = QtWidgets.QFrame()
		self.controls_frame.setLayout(QtWidgets.QGridLayout())
		self.controls_frame.layout().setContentsMargins(10, 10, 10, 10)
		
		self.scroll_area = VerticalScrollArea(self.controls_frame)
		
		self.layout().addWidget(self.scroll_area)
		
		self.button_frame = QtWidgets.QFrame()
		self.button_frame.setLayout(QtWidgets.QHBoxLayout())
		self.button_frame.layout().setContentsMargins(5, 5, 5, 5)
		self.button_frame.layout().addStretch()
		button_submit = QtWidgets.QPushButton("Submit")
		button_submit.clicked.connect(self.on_submit)
		self.button_frame.layout().addWidget(button_submit)
		button_reset = QtWidgets.QPushButton("Reset")
		button_reset.clicked.connect(self.on_reset)
		self.button_frame.layout().addWidget(button_reset)
		self.layout().addWidget(self.button_frame)
		
		for element in self.form_tool.elements:
			if issubclass(element.__class__, Select):
				self.add_select(element)
			elif isinstance(element, Unique):
				self.add_unique(element)
			elif isinstance(element, ColumnBreak):
				self.add_column()
			elif issubclass(element.__class__, UserControl):
				self.add_frame(element)
			elif issubclass(element.__class__, Group):
				self.add_group(element)
		
		self.adjust_labels()
	
	def set_enabled(self, state):
		
		if self.buttonBox is not None:
			self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(state)
	
	def add_column(self):
		
		self.columns.append(DialogColumn())
		self.controls_frame.layout().addWidget(self.columns[-1], 0, len(self.columns) - 1)
	
	def add_frame(self, user_control):
		
		self.columns[-1].add_widget(DialogFrame(self._vusertools._cmodel, user_control))
	
	def add_group(self, user_group):
		
		dialog_group = None
		if isinstance(user_group, MultiGroup):
			dialog_group = DialogMultiGroup(self._vusertools._cmodel, user_group)
			dialog_group.entry_added.connect(self.on_entry_added)
			dialog_group.entry_removed.connect(self.on_entry_removed)
		elif isinstance(user_group, Group):
			dialog_group = DialogGroup(self._vusertools._cmodel, user_group)
			dialog_group.entry_removed.connect(self.on_entry_removed)
		if dialog_group is None:
			return
		self.columns[-1].add_widget(dialog_group)
	
	def add_select(self, user_select):
	
		self.selects.append([user_select.dclass, user_select.descriptor])
	
	def add_unique(self, user_unique):
		
		self.unique.add(user_unique.dclass)
	
	def multigroups(self):
		# returns [DialogMultiGroup(), ...]
		
		groups = []
		for column in self.columns:
			for group in column.findChildren(DialogMultiGroup, options = QtCore.Qt.FindDirectChildrenOnly):
				groups.append(group)
		return groups
	
	def frames(self):
		# returns frames, framesets
		# frames = [DialogFrame(), ...]
		# framesets = [[DialogFrame(), ...], ...]
		
		frames, framesets = [], []
		for column in self.columns:
			for element in column.findChildren(QtWidgets.QWidget, options = QtCore.Qt.FindDirectChildrenOnly):
				if isinstance(element, DialogGroup):
					frames += element.frames()
				elif isinstance(element, DialogMultiGroup):
					framesets += element.framesets()
				elif isinstance(element, DialogFrame):
					frames.append(element)
		return frames, framesets
	
	def adjust_labels(self):
		
		for column in self.columns:
			wmax = 0
			for label in column.findChildren(QtWidgets.QLabel):
				wmax = max(wmax, label.sizeHint().width())
			for label in column.findChildren(QtWidgets.QLabel):
				label.setFixedWidth(wmax)
				label.setAlignment(QtCore.Qt.AlignRight)
	
	def update_lookups(self):
		
		for column in self.columns:
			for element in column.findChildren(QtWidgets.QWidget):
				if isinstance(element, DialogGroup) or isinstance(element, DialogFrame):
					element.populate_lookup()
	
	def sizeHint(self):
		
		margins = self.controls_frame.layout().contentsMargins()
		w = self.controls_frame.sizeHint().width() + self.scroll_area.verticalScrollBar().width() + margins.left() + margins.right()
		h = self.controls_frame.sizeHint().height() + self.button_frame.sizeHint().height() + margins.top() + margins.bottom()
		return QtCore.QSize(w, h)
	
	def resizeEvent(self, event):
		
		if not self.columns:
			return
		w = 0
		for column in self.columns:
			w += column.sizeHint().width()
		w = int(round(w / len(self.columns)))
		for col in range(len(self.columns)):
			self.controls_frame.layout().setColumnMinimumWidth(col, w)
	
	@QtCore.Slot()
	def on_entry_added(self):
		
		self.adjust_labels()
	
	@QtCore.Slot(list)
	def on_entry_removed(self, obj_ids):
		
		pass
	
	@QtCore.Slot()
	def on_submit(self):
		
		pass
	
	@QtCore.Slot()
	def on_reset(self):
		
		pass
	
	@QtCore.Slot()
	def on_data_changed(self):
		
		self.update_lookups()

class VerticalScrollArea(QtWidgets.QScrollArea):
	
	def __init__(self, contents):
		
		self.contents = contents
		
		QtWidgets.QScrollArea.__init__(self)
		
		self.setWidgetResizable(True)
		self.setFrameStyle(QtWidgets.QFrame.NoFrame)
		self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
		self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
		self.contents.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
		self.setWidget(self.contents)
	
	def resizeEvent(self, event):
		
		self.contents.setMinimumWidth(self.width() - self.verticalScrollBar().width())

